"""
Exit signals using indicators.

"""

import random
import numpy as np
from technicalmethods.methods import Indicators

class IndicatorExit():
    """
    Calculate exit signals using indicators.

    """

    @staticmethod
    def exit_parabolic_sar(
            prices, time_period, acceleration_factor, sip_price):
        """
        Calculate exit based on a Parabolic SAR.

        Parameters
        ----------
        prices : DataFrame
            The OHLC data
        trade_number : Series
            Array of trade numbers.
        end_of_day_position : Series
            The close of day, long/short/flat position.
        time_period : Int
            The length of time for the initial rolling lookback.
        acceleration_factor : Float
            The acceleration factor to use.
        sip_price : Bool
            Whether to set the SIP of the Parabolic SAR exit to n-day
            high / low or to the high of the previous trade. The default is
            False.

        Returns
        -------
        prices : DataFrame
            The OHLC data
        parabolic_sar_exit : Series
            The exit signals.

        """

        trade_number = prices['raw_trade_number']
        end_of_day_position = prices['raw_end_of_day_position']

        # Extract high, low and close series from the DataFrame
        high = prices['High']
        low = prices['Low']

        # Calculate rolling min and max closing prices based on time period
        rolling_high = prices['High'].rolling(time_period).max()
        rolling_low = prices['Low'].rolling(time_period).min()

        # Initialize zero arrays to store data
        sar = np.array([0.0]*len(prices))
        parabolic_sar_exit = np.array([0]*len(prices))
        extreme_price = np.array([0.0]*len(prices))
        incremental_af = np.array([0.0]*len(prices))
        ep_sar_diff = np.array([0.0]*len(prices))
        af_x_diff = np.array([0.0]*len(prices))

        # For each row in the data
        for row in range(1, len(prices)):

            # If there is a trade on
            if trade_number[row] != 0:

                # Find the row that relates to the trade entry
                trade_first_row = prices.index.get_loc(
                    prices[trade_number==trade_number[row]].index[0])
                trade_row_num = row - trade_first_row

                # If it is the trade entry day
                if trade_row_num == 0:

                    # If it is the first trade in the data
                    if trade_number[row] == 1:

                        # If there is a long position
                        if end_of_day_position[row] > 0:

                            # Set the initial point to the n-day low
                            initial_point = rolling_low[row]

                            # Set the extreme price to the day's high
                            extreme_price[row] = high[row]

                        # If there is a short position
                        elif end_of_day_position[row] < 0:

                            # Set the initial point to the n-day high
                            initial_point = rolling_high[row]

                            # Set the extreme price to the day's low
                            extreme_price[row] = low[row]

                    # For every other trade
                    else:
                        # If there is a long position
                        if end_of_day_position[row] > 0:

                            # If the sip_price flag is True
                            if sip_price:

                                # Set the initial point to the n-day low close
                                initial_point = rolling_low[row]

                            # If the sip_price flag is False
                            else:
                                # Set the initial point to the previous trade
                                # low
                                initial_point = min(
                                    prices[trade_number==trade_number[row]-1][
                                        'Low'])

                            # Set the extreme price to the day's high
                            extreme_price[row] = high[row]

                        # If there is a short position
                        elif end_of_day_position[row] < 0:

                            # If the sip_price flag is True
                            if sip_price:

                                # Set the initial point to the n-day low close
                                initial_point = rolling_high[row]

                            # If the sip_price flag is False
                            else:
                                # Set the initial point to the previous trade
                                # high
                                initial_point = max(
                                    prices[trade_number==trade_number[row]-1][
                                        'High'])

                                # Set the extreme price to the day's high
                                extreme_price[row] = low[row]

                        # Set the sar to the initial point
                        sar[row] = initial_point

                        # Set the dynamic incremental_af to the initial
                        # acceleration factor
                        incremental_af[row] = acceleration_factor

                # If it is not the trade entry day
                else:
                    # If the previous day was long
                    if end_of_day_position[row-1] == 1:

                        # If the previous day's sar was greater than the
                        # previous day's
                        # low
                        if sar[row-1] > low[row-1]:

                            # Set the signal to exit
                            parabolic_sar_exit[row-1] = -1

                            # Set the new sar to the previous trades extreme
                            # price
                            sar[row] = extreme_price[row-1]

                        # If the previous day's sar plus the acceleration
                        # factor multiplied by the difference between the
                        # extreme price and the sar is greater than the lowest
                        # low of the previous 2 days
                        elif (sar[row-1] + af_x_diff[row-1]
                            > min(low[row-1], low[row-2])):

                            # Set the sar to the lowest low of the previous 2
                            # days
                            sar[row] = min(low[row-1], low[row-2])

                        # Otherwise
                        else:
                            # Set the sar to the previous day's sar plus the
                            # acceleration factor multiplied by the difference
                            # between the extreme price and the sar
                            sar[row] = sar[row-1] + af_x_diff[row-1]

                    # Otherwise if the previous day was short
                    elif end_of_day_position[row-1] == -1:
                        # If the previous day's sar was less than the previous
                        # day's high
                        if sar[row-1] < high[row-1]:

                            # Set the signal to exit
                            parabolic_sar_exit[row-1] = 1

                            # Set the new sar to the previous trades extreme
                            # price
                            sar[row] = extreme_price[row-1]

                        # If the previous day's sar less the acceleration
                        # factor multiplied by the difference between the
                        # extreme price and the sar is less than the highest
                        # high of the previous 2 days
                        elif (sar[row-1] - af_x_diff[row-1]
                            < max(high[row-1], high[row-2])):

                            # Set the sar to the highest high of the previous
                            # 2 days
                            sar[row] = max(high[row-1], high[row-2])

                        # Otherwise
                        else:
                            # Set the sar to the previous day's sar minus the
                            # acceleration factor multiplied by the difference
                            # between the extreme price and the sar
                            sar[row] = sar[row-1] - af_x_diff[row-1]

                # If the current trade direction is long
                if end_of_day_position[row] == 1:

                    # Set the extreme price to the greater of the previous
                    # day's extreme price and the current day's high
                    extreme_price[row] = max(extreme_price[row-1], high[row])

                    # If the trade is making a new high
                    if extreme_price[row] > extreme_price[row-1]:

                        # Increment the acceleration factor by the input value
                        # to a max of 0.2
                        incremental_af[row] = min(
                            incremental_af[row-1] + acceleration_factor, 0.2)

                    # Otherwise
                    else:
                        # Set the acceleration factor to the same as the
                        # previous day
                        incremental_af[row] = incremental_af[row-1]

                # Otherwise if the current trade direction is short
                else:
                    # Set the extreme price to the lesser of the previous day's
                    # extreme price and the current day's low
                    extreme_price[row] = min(extreme_price[row-1], low[row])

                    # If the trade is making a new low
                    if extreme_price[row] < extreme_price[row-1]:

                        # Increment the acceleration factor by the input value
                        # to a max of 0.2
                        incremental_af[row] = min(
                            incremental_af[row-1] + acceleration_factor, 0.2)

                    # Otherwise
                    else:
                        # Set the acceleration factor to the same as the
                        # previous day
                        incremental_af[row] = incremental_af[row-1]

            # Calculate the absolute value of the difference between the
            # extreme price and the sar
            ep_sar_diff[row] = abs(sar[row] - extreme_price[row])

            # Calculate the difference between the extreme price and the sar
            # multiplied by the acceleration factor
            af_x_diff[row] =  ep_sar_diff[row] * incremental_af[row]

        # Set the DataFrame columns to the numpy arrays
        prices['rolling_high_sar_exit'] = rolling_high
        prices['rolling_low_sar_exit'] = rolling_low
        prices['sar_exit'] = sar
        prices['ep_sar_exit'] = extreme_price
        prices['af_sar_exit'] = incremental_af
        prices['ep_sar_diff_exit'] = ep_sar_diff
        prices['af_x_diff_exit'] = af_x_diff

        return prices, parabolic_sar_exit


    @staticmethod
    def exit_rsi_trail(prices, time_period, oversold, overbought):
        """
        Calculate exit based on a trailing RSI - a down close when the RSI is
        overbought when long or an up close when the RSI is oversold when
        short.

        Parameters
        ----------
        prices : DataFrame
            The OHLC data.
        trade_number : Series
            Array of trade numbers.
        end_of_day_position : Series
            The close of day, long/short/flat position.
        time_period : Int
            The length of time for the RSI calculation.
        oversold : Int
            The oversold level to use.
        overbought : Int
            The overbought level to use.

        Returns
        -------
        prices : DataFrame
            The OHLC data.
        rsi_trail_exit : Series
            The exit signals.

        """

        trade_number = prices['raw_trade_number']
        end_of_day_position = prices['raw_end_of_day_position']

        # Calculate RSI
        rsi = Indicators.RSI(close=prices['Close'], time_period=time_period)

        # Create an empty array to store the signals
        rsi_trail_exit = np.array([0]*len(prices))

        # For each row in the data
        for row in range(1, len(prices)):

            # If there is a trade on
            if trade_number[row] != 0:

                # If there is a long position
                if end_of_day_position[row] > 0:

                    # If todays close is less than the previous days close
                    # and todays RSI is greater than the overbought level
                    if (prices['Close'][row] < prices['Close'][row-1]
                        and rsi[row] > overbought):

                        # Set the exit signal to -1
                        rsi_trail_exit[row] = -1

                # If there is a short position
                elif end_of_day_position[row] < 0:

                    # If todays close is greater than the previous days close
                    # and todays RSI is less than the oversold level
                    if (prices['Close'][row] > prices['Close'][row-1]
                        and rsi[row] < oversold):

                        # Set the exit signal to 1
                        rsi_trail_exit[row] = 1
                else:
                    # Set the exit signal to 0
                    rsi_trail_exit[row] = 0

        # Set the DataFrame column to the numpy array
        prices['RSI_exit'] = rsi

        return prices, rsi_trail_exit


    @staticmethod
    def exit_key_reversal(prices, time_period):
        """
        Calculate exit based on a key reversal day - a new high combined with
        a down close when long or a new low combined with an up close when
        short.

        Parameters
        ----------
        prices : DataFrame
            The OHLC data.
        trade_number : Series
            Array of trade numbers.
        end_of_day_position : Series
            The close of day, long/short/flat position.
        time_period : Int
            The length of time for the rolling lookback calculation.

        Returns
        -------
        prices : DataFrame
            The OHLC data.
        key_reversal_exit : Series
            The exit signals.

        """

        trade_number = prices['raw_trade_number']
        end_of_day_position = prices['raw_end_of_day_position']

        # Calculate rolling high and low prices based on time period
        rolling_high = prices['High'].rolling(time_period).max()
        rolling_low = prices['Low'].rolling(time_period).min()

        # Create an empty array to store the signals
        key_reversal_exit = np.array([0]*len(prices))

        # For each row in the data
        for row in range(1, len(prices)):

            # If there is a trade on
            if trade_number[row] != 0:

                # If there is a long position
                if end_of_day_position[row] > 0:

                    # If the n-day high today is greater than yesterdays n-day
                    # high and todays close is less than the previous days
                    # close
                    if (rolling_high[row] > rolling_high[row-1]
                        and prices['Close'][row] < prices['Close'][row-1]):

                        # Set the exit signal to -1
                        key_reversal_exit[row] = -1

                # If there is a short position
                elif end_of_day_position[row] < 0:

                    # If the n-day low today is less than yesterdays n-day
                    # low and todays close is greater than the previous days
                    # close
                    if (rolling_low[row] < rolling_low[row-1]
                        and prices['Close'][row] > prices['Close'][row-1]):

                        # Set the exit signal to -1
                        key_reversal_exit[row] = 1

                else:
                    # Set the exit signal to 0
                    key_reversal_exit[row] = 0

        # Set the DataFrame columns to the numpy arrays
        prices['rolling_high_key'] = rolling_high
        prices['rolling_low_key'] = rolling_low

        return prices, key_reversal_exit


    @staticmethod
    def exit_volatility(prices, time_period, threshold):
        """
        Calculate exit based on an increase in volatility - a fall in price
        greater than the ATR * Threshold when long or a rise in price greater
        than the ATR * Threshold when when short.

        Parameters
        ----------
        prices : DataFrame
            The OHLC data.
        trade_number : Series
            Array of trade numbers.
        end_of_day_position : Series
            The close of day, long/short/flat position.
        time_period : Int
            The length of time for the ATR calculation.
        threshold : Float
            The multiplier for the ATR used to trigger the exit.

        Returns
        -------
        prices : DataFrame
            The OHLC data
        volatility_exit : Series
            The exit signals.

        """

        trade_number = prices['raw_trade_number']
        end_of_day_position = prices['raw_end_of_day_position']

        # Create an empty array to store the signals
        volatility_exit = np.array([0]*len(prices))

        # Calculate ATR
        atr = Indicators.ATR(
            prices['High'], prices['Low'], prices['Close'], time_period)

        # For each row in the data
        for row in range(1, len(prices)):

            # If there is a trade on
            if trade_number[row] != 0:

                # If there is a long position
                if end_of_day_position[row] > 0:

                    # If the decrease in closing price from yesterday to today
                    # is greater than the ATR * Threshold
                    if ((prices['Close'][row] - prices['Close'][row-1])
                        > (atr[row] * threshold)):

                        # Set the exit signal to -1
                        volatility_exit[row] = -1

                # If there is a short position
                elif end_of_day_position[row] < 0:

                    # If the increase in closing price from yesterday to today
                    # is greater than the ATR * Threshold
                    if ((prices['Close'][row-1] - prices['Close'][row])
                        > (atr[row] * threshold)):

                        # Set the exit signal to 1
                        volatility_exit[row] = 1

                else:
                    # Set the exit signal to 0
                    volatility_exit[row] = 0

        # Set the DataFrame column to the numpy array
        prices['ATR_exit'] = atr

        return prices, volatility_exit


    @staticmethod
    def exit_stochastic_crossover(prices, time_period):
        """
        Calculate exit based on a stochastic crossover - if the slow k crosses
        below the slow d when long or if the slow k crosses above the slow d
        when short.

        Parameters
        ----------
        prices : DataFrame
            The OHLC data.
        trade_number : Series
            Array of trade numbers.
        end_of_day_position : Series
            The close of day, long/short/flat position.
        time_period : Int
            The length of time for the fast k calculation.

        Returns
        -------
        prices : DataFrame
            The OHLC data
        stoch_cross_exit : Series
            The exit signals.

        """

        trade_number = prices['raw_trade_number']
        end_of_day_position = prices['raw_end_of_day_position']

        # Calculate slow k  and slow d
        slow_k, slow_d = Indicators.stochastic(
            prices['High'], prices['Low'], prices['Close'],
            fast_k_period=time_period, fast_d_period=3, slow_k_period=3,
            slow_d_period=3, output_type='slow')

        # Create an empty array to store the signals
        stoch_cross_exit = np.array([0]*len(prices))

        # For each row in the data
        for row in range(1, len(prices)):

            # If there is a trade on
            if trade_number[row] != 0:

                # If there is a long position
                if end_of_day_position[row] > 0:

                    # If the slow k crosses below the slow d
                    if (slow_k[row] < slow_d[row]
                        and slow_k[row-1] > slow_d[row-1]):

                        # Set the exit signal to -1
                        stoch_cross_exit[row] = -1

                # If there is a short position
                elif end_of_day_position[row] < 0:

                    # If the slow k crosses above the slow d
                    if (slow_k[row] > slow_d[row]
                        and slow_k[row-1] < slow_d[row-1]):

                        # Set the exit signal to 1
                        stoch_cross_exit[row] = 1

                else:
                    # Set the exit signal to 0
                    stoch_cross_exit[row] = 0

        # Set the DataFrame columns to the numpy arrays
        prices['slow_k_exit'] = slow_k
        prices['slow_d_exit'] = slow_d

        return prices, stoch_cross_exit


    @staticmethod
    def exit_random(prices):
        """
        Calculate exit based on the first losing day after a random time
        interval.

        Parameters
        ----------
        prices : DataFrame
            The OHLC data.
        trade_number : Series
            Array of trade numbers.
        end_of_day_position : Series
            The close of day, long/short/flat position.

        Returns
        -------
        prices : DataFrame
            The OHLC data
        random_exit : Series
            The exit signals.

        """

        trade_number = prices['raw_trade_number']
        end_of_day_position = prices['raw_end_of_day_position']

        # Generate a random number of days between 5 and 20
        exit_days = random.randint(5,20)

        # Create an empty array to store the signals
        random_exit = np.array([0]*len(prices))

        # For each row in the data
        for row in range(1, len(prices)):

            # Find the row that relates to the trade entry
            trade_first_row = prices.index.get_loc(
                    prices[trade_number==trade_number[row]].index[0])
            trade_row_num = row - trade_first_row

            # If there is a trade on
            if trade_number[row] != 0:

                # If the trade has been on for the random number of days
                if trade_row_num > exit_days-1:

                    # If there is a long position
                    if end_of_day_position[row] > 0:

                        # If todays close is less than the previous days close
                        if prices['Close'][row] < prices['Close'][row-1]:

                            # Set the exit signal to -1
                            random_exit[row] = -1

                    # If there is a short position
                    elif end_of_day_position[row] < 0:

                        # If todays close is greater than the previous days
                        # close
                        if prices['Close'][row] > prices['Close'][row-1]:

                            # Set the exit signal to 1
                            random_exit[row] = 1

                    else:
                        # Set the exit signal to 0
                        random_exit[row] = 0

        # Set the DataFrame column to the numpy array
        prices['random_days_exit'] = exit_days

        return prices, random_exit


    @staticmethod
    def exit_support_resistance(prices, time_period):
        """
        Calculate exit based on an n-day high / low.

        Parameters
        ----------
        prices : DataFrame
            The OHLC data.
        trade_number : Series
            Array of trade numbers.
        end_of_day_position : Series
            The close of day, long/short/flat position.
        time_period : Int
            The length of time for the indicator.

        Returns
        -------
        prices : DataFrame
            The OHLC data.
        support_resistance_exit : Series
            The exit signals.

        """

        trade_number = prices['raw_trade_number']
        end_of_day_position = prices['raw_end_of_day_position']

        # Calculate rolling min and max closing prices based on time period
        rolling_high_close = prices['Close'].rolling(time_period).max()
        rolling_low_close = prices['Close'].rolling(time_period).min()

        # Create empty arrays to store the signals
        support_resistance_exit = np.array([0.0]*len(prices))
        exit_level = np.array([0.0]*len(prices))

        # For each row in the data
        for row in range(1, len(prices)):

            # If there is a trade on
            if trade_number[row] != 0:

                # If there is a long position
                if end_of_day_position[row] > 0:

                    # Set the exit level to the n-day low close
                    exit_level[row] = rolling_low_close[row]

                    # If the close is greater than the exit level
                    if prices['Close'][row] < exit_level[row]:

                        # Set the exit signal to -1
                        support_resistance_exit[row] = -1

                # If there is a short position
                elif end_of_day_position[row] < 0:

                    # Set the exit level to the n-day high close
                    exit_level[row] = rolling_high_close[row]

                    # If the close is greater than the exit level
                    if prices['Close'][row] > exit_level[row]:

                        # Set the exit signal to 1
                        support_resistance_exit[row] = 1

                else:
                    # Set the exit signal to 0
                    support_resistance_exit[row] = 0

        # Set the DataFrame columns to the numpy arrays
        prices['rolling_high_close_sr_exit'] = rolling_high_close
        prices['rolling_low_close_sr_exit'] = rolling_low_close
        prices['exit_level'] = exit_level

        return prices, support_resistance_exit


    @staticmethod
    def exit_immediate_profit(prices, time_period):
        """
        Calculate exit based on an immediate n-day profit.

        Parameters
        ----------
        prices : DataFrame
            The OHLC data.
        trade_number : Series
            Array of trade numbers.
        end_of_day_position : Series
            The close of day, long/short/flat position.
        time_period : Int
            The length of time for the indicator.

        Returns
        -------
        prices : DataFrame
            The OHLC data.
        immediate_profit_exit : Series
            The exit signals.
        """

        trade_number = prices['raw_trade_number']
        end_of_day_position = prices['raw_end_of_day_position']

        # Create an empty array to store the signals
        immediate_profit_exit = np.array([0.0]*len(prices))

        # For each row in the data
        for row in range(1, len(prices)):

            # If there is a trade on
            if trade_number[row] != 0:

                # Find the row that relates to the trade entry
                trade_first_row = prices.index.get_loc(
                    prices[trade_number==trade_number[row]].index[0])
                trade_row_num = row - trade_first_row

                # After the given number of days
                if trade_row_num == time_period-1:

                    # If there is a long position
                    if end_of_day_position[row] > 0:

                        # If the trade is losing money
                        if (prices['Close'][row]
                            < prices['Close'][row-time_period]):

                            # Set the exit signal to -1
                            immediate_profit_exit[row] = -1

                    # If there is a short position
                    elif end_of_day_position[row] < 0:

                        # If the trade is losing money
                        if (prices['Close'][row]
                            > prices['Close'][row-time_period]):

                            # Set the exit signal to 1
                            immediate_profit_exit[row] = 1

                    else:
                        # Set the exit signal to 0
                        immediate_profit_exit[row] = 0

        return prices, immediate_profit_exit


    @staticmethod
    def exit_nday_range(prices, time_period):
        """
        Calculate exit based on an n-day range.

        Parameters
        ----------
        prices : DataFrame
            The OHLC data.
        trade_number : Series
            Array of trade numbers.
        end_of_day_position : Series
            The close of day, long/short/flat position.
        time_period : Int
            The length of time for the indicator.

        Returns
        -------
        prices : DataFrame
            The OHLC data.
        nday_range_exit : Series
            The exit signals.

        """

        trade_number = prices['raw_trade_number']
        end_of_day_position = prices['raw_end_of_day_position']

        # The highest high minus the lowest low of the last n days
        n_day_low = prices['Low'].rolling(time_period).min()
        n_day_high = prices['High'].rolling(time_period).max()
        high_low_range = n_day_high - n_day_low

        # The largest high low daily range of the last n bars
        bar_range = (prices['High'] - prices['Low']).rolling(time_period).max()

        # Create an empty array to store the signals
        nday_range_exit = np.array([0.0]*len(prices))

        # For each row in the data
        for row in range(1, len(prices)):

            # If there is a trade on
            if trade_number[row] != 0:

                # Find the row that relates to the trade entry
                trade_first_row = prices.index.get_loc(
                    prices[trade_number==trade_number[row]].index[0])
                trade_row_num = row - trade_first_row

                # If it is the trade entry date
                if trade_row_num == 0:

                    # Set flags for if the targets have been hit
                    pt_1_hit = False
                    pt_2_hit = False

                    # Set target 1 as the lower of the high low range of the
                    # last n days and the longest bar of the last n days
                    target_1 = min(high_low_range[row], bar_range[row])
                    target_2 = max(high_low_range[row], bar_range[row])

                    # If the position is long, add these to the close to set
                    # the price targets
                    if end_of_day_position[row] > 0:
                        price_target_1 = prices['Close'][row] + target_1
                        price_target_2 = prices['Close'][row] + target_2

                    # Otherwise subtract from the price
                    else:
                        price_target_1 = prices['Close'][row] - target_1
                        price_target_2 = prices['Close'][row] - target_2

                # For every other day in the trade
                else:

                    # If the position is long
                    if end_of_day_position[row] > 0:

                        # If the close is above the target 1 and this has not
                        # yet been hit
                        if (prices['Close'][row] > price_target_1
                            and pt_1_hit is False):

                            # Set the exit signal to -1
                            nday_range_exit[row] = -1

                            # Set the profit target 1 hit flag to True
                            pt_1_hit = True

                        # If the close is above the target 2 and this has not
                        # yet been hit
                        elif (prices['Close'][row] > price_target_2
                              and pt_2_hit is False):

                            # Set the exit signal to -1
                            nday_range_exit[row] = -1

                            # Set the profit target 2 hit flag to True
                            pt_2_hit = True

                        # Otherwise
                        else:
                            # Set the exit signal to 0
                            nday_range_exit[row] = 0

                    # If the position is short
                    else:

                        # If the close is below the target 1 and this has not
                        # yet been hit
                        if (prices['Close'][row] < price_target_1
                            and pt_1_hit is False):

                            # Set the exit signal to 1
                            nday_range_exit[row] = 1

                            # Set the profit target 1 hit flag to True
                            pt_1_hit = True

                        # If the close is above the target 2 and this has not
                        # yet been hit
                        elif (prices['Close'][row] < price_target_2
                              and pt_2_hit is False):

                            # Set the exit signal to 1
                            nday_range_exit[row] = 1

                            # Set the profit target 2 hit flag to True
                            pt_2_hit = True

                        # Otherwise
                        else:
                            # Set the exit signal to 0
                            nday_range_exit[row] = 0

        return prices, nday_range_exit



"""
Indicator Entry signals

"""

import numpy as np
from technicalmethods.methods import Indicators

class IndicatorEntry():
    """
    Calculate entry signals using indicators / oscillators.

    """

    @staticmethod
    def entry_parabolic_sar(prices, acceleration_factor):
        """
        Entry signals based on Parabolic SAR

        Parameters
        ----------
        prices : DataFrame
            The OHLC data
        acceleration_factor : Float
            The acceleration factor to use.

        Returns
        -------
        prices : DataFrame
            The OHLC data with additional columns.
        start : Int
            The first valid date row to calculate from.
        trade_signal : Series
            The series of Buy / Sell signals

        """
        # Extract high, low and close series from the DataFrame
        high = prices['High']
        low = prices['Low']
        close = prices['Close']

        # Create empty arrays to store data
        sar = np.array([0.0]*len(prices))
        direction = np.array([0.0]*len(prices))
        trade_signal = np.array([0.0]*len(prices))
        extreme_price = np.array([0.0]*len(prices))
        incremental_af = np.array([0.0]*len(prices))
        ep_sar_diff = np.array([0.0]*len(prices))
        af_x_diff = np.array([0.0]*len(prices))

        # Configure initial values
        initial_sip = 0.975 * close[0]
        sar[0] = initial_sip
        direction[0] = 1
        extreme_price[0] = high[0]
        incremental_af[0] = 0.02
        ep_sar_diff[0] = abs(sar[0] - extreme_price[0])
        af_x_diff[0] =  ep_sar_diff[0] * incremental_af[0]
        start = 1
        init_flag = True

        # Loop through, starting from the second row
        for row in range(start, len(prices)):

            # If the previous day was long
            if direction[row-1] == 1:

                # If the previous day's sar was greater than the previous day's
                # low
                if sar[row-1] > low[row-1]:

                    # If this is the starting trade
                    if init_flag:

                        # Enter short
                        trade_signal[row-1] = -1

                        # Set the flag to False
                        init_flag = False

                    else:
                        # Close the long position and go short
                        trade_signal[row-1] = -2

                    # Set the new sar to the previous trades extreme price
                    sar[row] = extreme_price[row-1]

                    # Switch the trade direction from long to short
                    direction[row] = -direction[row-1]

                # If the previous day's sar plus the acceleration factor
                # multiplied by the difference between the extreme price and
                # the sar is greater than the lowest low of the previous 2 days
                elif (sar[row-1] + af_x_diff[row-1]
                      > min(low[row-1], low[row-2])):

                    # Set the sar to the lowest low of the previous 2 days
                    sar[row] = min(low[row-1], low[row-2])

                    # Set the direction to the same as the previous day
                    direction[row] = direction[row-1]

                # Otherwise
                else:
                    # Set the sar to the previous day's sar plus the
                    # acceleration factor multiplied by the difference between
                    # the extreme price and the sar
                    sar[row] = sar[row-1] + af_x_diff[row-1]

                    # Set the direction to the same as the previous day
                    direction[row] = direction[row-1]

            # Otherwise if the previous day was short
            else:
                # If the previous day's sar was less than the previous day's
                # high
                if sar[row-1] < high[row-1]:

                    # Close the short position and go long
                    trade_signal[row-1] = 2

                    # Set the new sar to the previous trades extreme price
                    sar[row] = extreme_price[row-1]

                    # Switch the trade direction from short to long
                    direction[row] = -direction[row-1]

                # If the previous day's sar less the acceleration factor
                # multiplied by the difference between the extreme price and
                # the sar is less than the highest high of the previous 2 days
                elif (sar[row-1] - af_x_diff[row-1]
                      < max(high[row-1], high[row-2])):

                    # Set the sar to the highest high of the previous 2 days
                    sar[row] = max(high[row-1], high[row-2])

                    # Set the direction to the same as the previous day
                    direction[row] = direction[row-1]

                # Otherwise
                else:
                    # Set the sar to the previous day's sar minus the
                    # acceleration factor multiplied by the difference between
                    # the extreme price and the sar
                    sar[row] = sar[row-1] - af_x_diff[row-1]

                    # Set the direction to the same as the previous day
                    direction[row] = direction[row-1]

            # If the current trade direction is long
            if direction[row] == 1:

                # If the trade has just reversed direction
                if direction[row] != direction[row-1]:

                    # Set the extreme price to the day's high
                    extreme_price[row] = high[row]

                    # Set the initial acceleration factor to the input value
                    incremental_af[row] = acceleration_factor

                # If the trade is the same as the previous day
                else:

                    # Set the extreme price to the greater of the previous
                    # day's extreme price and the current day's high
                    extreme_price[row] = max(extreme_price[row-1], high[row])

                    # If the trade is making a new high
                    if extreme_price[row] > extreme_price[row-1]:

                        # Increment the acceleration factor by the input value
                        # to a max of 0.2
                        incremental_af[row] = min(
                            incremental_af[row-1] + acceleration_factor, 0.2)

                    # Otherwise
                    else:

                        # Set the acceleration factor to the same as the
                        # previous day
                        incremental_af[row] = incremental_af[row-1]

            # Otherwise if the current trade direction is short
            else:

                # If the trade has just reversed direction
                if direction[row] != direction[row-1]:

                    # Set the extreme price to the day's low
                    extreme_price[row] = low[row]

                    # Set the initial acceleration factor to the input value
                    incremental_af[row] = acceleration_factor

                # If the trade is the same as the previous day
                else:

                    # Set the extreme price to the lesser of the previous day's
                    # extreme price and the current day's low
                    extreme_price[row] = min(extreme_price[row-1], low[row])

                    # If the trade is making a new low
                    if extreme_price[row] < extreme_price[row-1]:

                        # Increment the acceleration factor by the input value
                        # to a max of 0.2
                        incremental_af[row] = min(
                            incremental_af[row-1] + acceleration_factor, 0.2)

                    # Otherwise
                    else:

                        # Set the acceleration factor to the same as the
                        # previous day
                        incremental_af[row] = incremental_af[row-1]

            # Calculate the absolute value of the difference between the
            # extreme price and the sar
            ep_sar_diff[row] = abs(sar[row] - extreme_price[row])

            # Calculate the difference between the extreme price and the sar
            # multiplied by the acceleration factor
            af_x_diff[row] =  ep_sar_diff[row] * incremental_af[row]

        # Assign the series to the OHLC data
        prices['sar_entry'] = sar
        prices['direction_sar_entry'] = direction
        prices['ep_sar_entry'] = extreme_price
        prices['af_sar_entry'] = incremental_af
        prices['ep_sar_diff_entry'] = ep_sar_diff
        prices['af_x_diff_entry'] = af_x_diff

        return prices, start, trade_signal


    @staticmethod
    def entry_channel_breakout(prices, time_period):
        """
        Entry signals based on a channel breakout.

        Parameters
        ----------
        prices : DataFrame
            The OHLC data
        time_period : Int
            The number of days to use in the indicator calculation. The default
            is 20.

        Returns
        -------
        prices : DataFrame
            The OHLC data with additional columns.
        start : Int
            The first valid date row to calculate from.
        trade_signal : Series
            The series of Buy / Sell signals

        """
        # Calculate rolling min and max closing prices based on time period
        rolling_high_close = prices['Close'].rolling(time_period).max()
        rolling_low_close = prices['Close'].rolling(time_period).min()

        # Create start point based on lookback window
        start = np.where(~np.isnan(rolling_high_close))[0][0]

        # Create numpy array of zeros to store position signals
        position_signal = np.array([0]*len(rolling_high_close))

        # Create numpy array of zeros to store trade signals
        trade_signal = np.array([0]*len(rolling_high_close))

        # for each row in the DataFrame after the longest MA has started
        for row in range(start, len(rolling_high_close)):

            # If the price rises to equal or above the n-day high close
            if prices['Close'][row] >= rolling_high_close[row]:

                # Set the position signal to long
                position_signal[row] = 1

                # Signal to go long
                trade_signal[row] = 1 - position_signal[row-1]

            # If the price falls to equal or below the n-day low close
            elif prices['Close'][row] <= rolling_low_close[row]:

                # Set the position signal to short
                position_signal[row] = -1

                # Signal to go short
                trade_signal[row] = -1 - position_signal[row-1]

            # Otherwise, take no action
            else:
                trade_signal[row] = 0
                position_signal[row] = position_signal[row-1]

        # Assign the series to the OHLC data
        prices['rolling_high_close_entry'] = rolling_high_close
        prices['rolling_low_close_entry'] = rolling_low_close

        return prices, start, trade_signal


    @staticmethod
    def entry_stochastic_crossover(prices, time_period, oversold, overbought):
        """
        Entry signals based on slow k / slow d stochastics crossing.

        Parameters
        ----------
        prices : DataFrame
            The OHLC data
        time_period : Int
            The number of days to use in the indicator calculation.
        oversold : Int
            The oversold level to use.
        overbought : Int
            The overbought level to use.

        Returns
        -------
        prices : DataFrame
            The OHLC data with additional columns.
        start : Int
            The first valid date row to calculate from.
        trade_signal : Series
            The series of Buy / Sell signals

        """
        # Create Stochastics for the specified time period
        slow_k, slow_d = Indicators.stochastic(
            prices['High'], prices['Low'], prices['Close'],
            fast_k_period=time_period)

        # Create start point based on slow d
        start = np.where(~np.isnan(slow_d))[0][0]

        # Create numpy array of zeros to store position signals
        position_signal = np.array([0]*len(slow_d))

        # Create numpy array of zeros to store trade signals
        trade_signal = np.array([0]*len(slow_d))

        # for each row in the DataFrame after the slow d has started
        for row in range(start, len(slow_d)):

            # If the slow k crosses above the slow d from below having been
            # below the lower level
            if (slow_k[row] > slow_d[row]
                and slow_k[row-1] < slow_d[row-1]
                and slow_k[row] < oversold):

                # Set the position signal to long
                position_signal[row] = 1

                # Signal to go long
                trade_signal[row] = 1 - position_signal[row-1]

            # If the slow k crosses below the slow d from above having been
            # above the upper level
            elif (slow_k[row] < slow_d[row]
                and slow_k[row-1] > slow_d[row-1]
                and slow_k[row] > overbought):

                # Set the position signal to short
                position_signal[row] = -1

                # Signal to go short
                trade_signal[row] = -1 - position_signal[row-1]

            # Otherwise, take no action
            else:
                trade_signal[row] = 0
                position_signal[row] = position_signal[row-1]

        # Assign the series to the OHLC data
        prices['slow_k_entry'] = slow_k
        prices['slow_d_entry'] = slow_d

        return prices, start, trade_signal


    @staticmethod
    def entry_stochastic_over_under(prices, time_period, oversold, overbought):
        """
        Entry signals based on slow k / slow d stochastics crossing
        overbought / oversold levels.

        Parameters
        ----------
        prices : DataFrame
            The OHLC data
        time_period : Int
            The number of days to use in the indicator calculation.
        oversold : Int
            The oversold level to use.
        overbought : Int
            The overbought level to use.

        Returns
        -------
        prices : DataFrame
            The OHLC data with additional columns.
        start : Int
            The first valid date row to calculate from.
        trade_signal : Series
            The series of Buy / Sell signals

        """
        # Create Stochastics for the specified time period
        slow_k, slow_d = Indicators.stochastic(
            prices['High'], prices['Low'], prices['Close'],
            fast_k_period=time_period)

        # Create start point based on slow d
        start = np.where(~np.isnan(slow_d))[0][0]

        # Create numpy array of zeros to store position signals
        position_signal = np.array([0]*len(slow_d))

        # Create numpy array of zeros to store trade signals
        trade_signal = np.array([0]*len(slow_d))

        # for each row in the DataFrame after the slow d has started
        for row in range(start, len(slow_d)):

            # If both slow k and slow d cross above the lower level from below
            if (min(slow_k[row], slow_d[row])
                > oversold
                > min(slow_k[row-1], slow_d[row-1])):

                # Set the position signal to long
                position_signal[row] = 1

                # Signal to go long
                trade_signal[row] = 1 - position_signal[row-1]

            # If both slow k and slow d cross above the upper level from below
            elif (min(slow_k[row], slow_d[row])
                  > overbought
                  > min(slow_k[row-1], slow_d[row-1])):

                # Set the position signal to flat
                position_signal[row] = 0

                # Signal to go flat
                trade_signal[row] = 0 - position_signal[row-1]

            # If both slow k and slow d cross below the upper level from above
            elif (max(slow_k[row], slow_d[row])
                  < overbought
                  < max(slow_k[row-1], slow_d[row-1])):

                # Set the position signal to short
                position_signal[row] = -1

                # Signal to go short
                trade_signal[row] = -1 - position_signal[row-1]

            # If both slow k and slow d cross below the lower level from above
            elif (max(slow_k[row], slow_d[row])
                  < oversold
                  < max(slow_k[row-1], slow_d[row-1])):

                # Set the position signal to flat
                position_signal[row] = 0

                # Signal to go flat
                trade_signal[row] = 0 - position_signal[row-1]

            # Otherwise, take no action
            else:
                trade_signal[row] = 0
                position_signal[row] = position_signal[row-1]

        # Assign the series to the OHLC data
        prices['slow_k_entry'] = slow_k
        prices['slow_d_entry'] = slow_d

        return prices, start, trade_signal


    @staticmethod
    def entry_stochastic_pop(prices, time_period, oversold, overbought):
        """
        Entry signals based on the Stochastic Pop method

        Parameters
        ----------
        prices : DataFrame
            The OHLC data
        time_period : Int
            The number of days to use in the indicator calculation.
        oversold : Int
            The oversold level to use.
        overbought : Int
            The overbought level to use.

        Returns
        -------
        prices : DataFrame
            The OHLC data with additional columns.
        start : Int
            The first valid date row to calculate from.
        trade_signal : Series
            The series of Buy / Sell signals

        """
        # Create Stochastics for the specified time period
        slow_k, slow_d = Indicators.stochastic(
            prices['High'], prices['Low'], prices['Close'],
            fast_k_period=time_period)

        # Create start point based on slow d
        start = np.where(~np.isnan(slow_d))[0][0]

        # Create numpy array of zeros to store position signals
        position_signal = np.array([0]*len(slow_d))

        # Create numpy array of zeros to store trade signals
        trade_signal = np.array([0]*len(slow_d))

        # for each row in the DataFrame after the slow d has started
        for row in range(start, len(slow_d)):

            # If both slow k and slow d cross above the upper level from below
            if (min(slow_k[row], slow_d[row])
                > overbought
                > min(slow_k[row-1], slow_d[row-1])):

                # Set the position signal to long
                position_signal[row] = 1

                # Signal to go long
                trade_signal[row] = 1 - position_signal[row-1]

            # If slow k and slow d cross when the position is long
            elif ((position_signal[row] == 1)
                  and np.sign(slow_k[row] - slow_d[row]) != np.sign(
                      slow_k[row-1] - slow_d[row-1])):

                # Set the position signal to flat
                position_signal[row] = 0

                # Signal to go flat
                trade_signal[row] = 0 - position_signal[row-1]

            # If both slow k and slow d cross below the lower level from above
            elif (max(slow_k[row], slow_d[row])
                  < oversold
                  < max(slow_k[row-1], slow_d[row-1])):

                # Set the position signal to short
                position_signal[row] = -1

                # Signal to go short
                trade_signal[row] = -1 - position_signal[row-1]

            # If slow k and slow d cross when the position is short
            elif ((position_signal[row] == -1)
                  and np.sign(slow_k[row] - slow_d[row]) != np.sign(
                      slow_k[row-1] - slow_d[row-1])):

                # Set the position signal to flat
                position_signal[row] = 0

                # Signal to go flat
                trade_signal[row] = 0 - position_signal[row-1]

            # Otherwise, take no action
            else:
                trade_signal[row] = 0
                position_signal[row] = position_signal[row-1]

        # Assign the series to the OHLC data
        prices['slow_k_entry'] = slow_k
        prices['slow_d_entry'] = slow_d

        return prices, start, trade_signal


    @staticmethod
    def entry_rsi(prices, time_period, oversold, overbought):
        """
        Entry signals based on the Relative Strength Index

        Parameters
        ----------
        prices : DataFrame
            The OHLC data
        time_period : Int
            The number of days to use in the indicator calculation.
        oversold : Int
            The oversold level to use.
        overbought : Int
            The overbought level to use.

        Returns
        -------
        prices : DataFrame
            The OHLC data with additional columns.
        start : Int
            The first valid date row to calculate from.
        trade_signal : Series
            The series of Buy / Sell signals

        """
        # Create RSI for the specified time period
        rsi = Indicators.RSI(prices['Close'], time_period)

        # Create start point based on lookback window
        start = np.where(~np.isnan(rsi))[0][0]

        # Create numpy array of zeros to store position signals
        position_signal = np.array([0]*len(rsi))

        # Create numpy array of zeros to store trade signals
        trade_signal = np.array([0]*len(rsi))

        # for each row in the DataFrame after the cci has started
        for row in range(start, len(rsi)):

            # If the cci crosses above the threshold from below
            if rsi[row] < oversold < rsi[row-1]:

                # Set the position signal to long
                position_signal[row] = 1

                # Signal to go long
                trade_signal[row] = 1 - position_signal[row-1]

            # If the cci crosses below the threshold from above
            elif rsi[row] > overbought > rsi[row-1]:

                # Set the position signal to short
                position_signal[row] = -1

                # Signal to go short
                trade_signal[row] = -1 - position_signal[row-1]

            # Otherwise, take no action
            else:
                trade_signal[row] = 0
                position_signal[row] = position_signal[row-1]

        # Assign the series to the OHLC data
        prices['RSI_entry'] = rsi

        return prices, start, trade_signal


    @staticmethod
    def entry_commodity_channel_index(prices, time_period, threshold):
        """
        Entry signals based on the Commodity Channel Index

        Parameters
        ----------
        prices : DataFrame
            The OHLC data
        time_period : Int
            The number of days to use in the indicator calculation.
        threshold : TYPE
            DESCRIPTION.

        Returns
        -------
        prices : DataFrame
            The OHLC data with additional columns.
        start : Int
            The first valid date row to calculate from.
        trade_signal : Series
            The series of Buy / Sell signals

        """

        # Create CCI for the specified time period
        cci = Indicators.CCI(
            prices['High'], prices['Low'], prices['Close'], time_period)

        # Create start point based on lookback window
        start = np.where(~np.isnan(cci))[0][0]

        # Create numpy array of zeros to store position signals
        position_signal = np.array([0]*len(cci))

        # Create numpy array of zeros to store trade signals
        trade_signal = np.array([0]*len(cci))

        # for each row in the DataFrame after the cci has started
        for row in range(start, len(cci)):

            # If the cci crosses above the threshold from below
            if cci[row] > threshold > cci[row-1]:

                # Set the position signal to long
                position_signal[row] = 1

                # Signal to go long
                trade_signal[row] = 1 - position_signal[row-1]

            # If the cci crosses below the threshold from above
            elif cci[row] < threshold < cci[row-1]:

                # Set the position signal to short
                position_signal[row] = -1

                # Signal to go short
                trade_signal[row] = -1 - position_signal[row-1]

            # Otherwise, take no action
            else:
                trade_signal[row] = 0
                position_signal[row] = position_signal[row-1]

        # Assign the series to the OHLC data
        prices['CCI_entry'] = cci

        return prices, start, trade_signal


    @staticmethod
    def entry_momentum(prices, time_period, threshold):
        """
        Entry signals based on n-day momentum

        Parameters
        ----------
        prices : DataFrame
            The OHLC data
        time_period : Int
            The number of days to use in the indicator calculation. The default
            is 10.
        threshold : Float
            The threshold used for taking signals. The default is 0.

        Returns
        -------
        prices : DataFrame
            The OHLC data with additional columns.
        start : Int
            The first valid date row to calculate from.
        trade_signal : Series
            The series of Buy / Sell signals

        """
        # Calculate past close based on time period
        n_day_close = prices['Close'].shift(time_period)

        # Create start point based on lookback window
        start = np.where(~np.isnan(n_day_close))[0][0]

        # Create numpy array of zeros to store position signals
        position_signal = np.array([0]*len(n_day_close))

        # Create numpy array of zeros to store trade signals
        trade_signal = np.array([0]*len(n_day_close))

        # for each row in the DataFrame after the longest MA has started
        for row in range(start, len(n_day_close)):

            # If the price rises to equal or above the n-day high close
            if prices['Close'][row] >= n_day_close[row] + threshold:

                # Set the position signal to long
                position_signal[row] = 1

                # Signal to go long
                trade_signal[row] = 1 - position_signal[row-1]

            # If the price falls to equal or below the n-day low close
            elif prices['Close'][row] <= n_day_close[row] - threshold:

                # Set the position signal to short
                position_signal[row] = -1

                # Signal to go short
                trade_signal[row] = -1 - position_signal[row-1]

            # Otherwise, take no action
            else:
                trade_signal[row] = 0
                position_signal[row] = position_signal[row-1]

        # Assign the series to the OHLC data
        prices['n_day_close'] = n_day_close
        prices['momentum'] = prices['Close'] - prices['n_day_close']

        return prices, start, trade_signal


    @staticmethod
    def entry_volatility(prices, time_period, threshold):
        """
        Entry signals based on a volatility breakout.

        Parameters
        ----------
        prices : DataFrame
            The OHLC data
        time_period : Int
            The number of days to use in the indicator calculation.
        threshold : Float
            The threshold used for taking signals. The default is 1.5.

        Returns
        -------
        prices : DataFrame
            The OHLC data with additional columns.
        start : Int
            The first valid date row to calculate from.
        trade_signal : Series
            The series of Buy / Sell signals

        """
        # Create ATR for the specified time period
        atr = Indicators.ATR(
            prices['High'], prices['Low'], prices['Close'], time_period)

        # Create start point based on lookback window
        start = np.where(~np.isnan(atr))[0][0]

        # Create numpy array of zeros to store position signals
        position_signal = np.array([0]*len(atr))

        # Create numpy array of zeros to store trade signals
        trade_signal = np.array([0]*len(atr))

        # for each row in the DataFrame after the atr has started
        for row in range(start, len(atr)):

            # If the increase in closing price exceeds the atr * threshold
            if ((prices['Close'][row] - prices['Close'][row-1])
                > (atr[row] * threshold)):

                # Set the position signal to long
                position_signal[row] = 1

                # Signal to go long
                trade_signal[row] = 1 - position_signal[row-1]

            # If the decrease in closing price exceeds the atr * threshold
            elif ((prices['Close'][row-1] - prices['Close'][row])
                  > (atr[row] * threshold)):

                # Set the position signal to short
                position_signal[row] = -1

                # Signal to go short
                trade_signal[row] = -1 - position_signal[row-1]

            # Otherwise, take no action
            else:
                trade_signal[row] = 0
                position_signal[row] = position_signal[row-1]

        # Assign the series to the OHLC data
        prices['ATR_entry'] = atr

        return prices, start, trade_signal



"""
Dictionary containing all the default parameters

"""

system_params_dict = {
    'df_profit_factor':2,
    'df_profit_bars':2,
    'df_stop_bars':2,
    'df_exit_factor':2,
    'df_equity':100000.00,
    'df_ticker':'$SPX',
    'df_bench_ticker':'$SPX',
    'df_ccy_1':'GBP',
    'df_ccy_2':'USD',
    'df_asset_type':'fx',
    'df_start_date':None,
    'df_end_date':None,
    'df_lookback':750,
    'df_ma1':5,
    'df_ma2':12,
    'df_ma3':20,
    'df_ma4':40,
    'df_simple_ma':True,
    'df_fixed_pos_size':5,
    'df_atr_pos_size':14,
    'df_position_type':'atr',
    'df_margin_%':0.2,
    'df_ticker_source':'norgate',
    'df_bench_source':'norgate',
    'df_slippage':5.0,
    'df_commission':0.00,
    'df_riskfree':0.0025,
    'df_entry_type':'2ma',
    'df_exit_type':'trailing_stop',
    'df_stop_type':'initial_dollar',
    'df_entry_period':14,
    'df_exit_period':5,
    'df_stop_period':5,
    'df_entry_threshold':0,
    'df_exit_threshold':0,
    'df_exit_amount':3000.00,
    'df_stop_amount':1500.00,
    'df_entry_oversold':25,
    'df_entry_overbought':75,
    'df_exit_oversold':25,
    'df_exit_overbought':75,
    'df_entry_acceleration_factor':0.02,
    'df_exit_acceleration_factor':0.02,
    'df_sip_price':False,
    'df_signals':True,
    'df_api_key':'',
    'df_position_risk_bps':500,
    'df_equity_inv_perc':0.75,

    # Columns names used for strategy indicators
    'df_entry_signal_indicators':{
        '2ma':('ma_1','ma_2'),
        '3ma':('ma_1','ma_2', 'ma_3'),
        '4ma':('ma_1','ma_2', 'ma_3', 'ma_4'),
        'sar':'sar_entry',
        'channel_breakout':(
            'rolling_low_close_entry',
            'rolling_high_close_entry'),
        'stoch_cross':(
            'slow_k_entry', 'slow_d_entry'),
        'stoch_over_under':(
            'slow_k_entry', 'slow_d_entry'),
        'stoch_pop':(
            'slow_k_entry', 'slow_d_entry'),
        'rsi':'RSI_entry',
        'cci':'CCI_entry',
        'momentum':'momentum',
        'volatility':'ATR_entry'
        },

    # Signals requiring changes to default parameters
    'df_entry_signal_labels':{
        '2ma':('Double ','MA Crossover'),
        '3ma':('Triple ','MA Crossover'),
        '4ma':('Quad ','MA Crossover'),
        'sar':'Parabolic SAR',
        'channel_breakout':'Channel Breakout',
        'stoch_cross':'Stochastic Crossover',
        'stoch_over_under':'Stochastic Over Under',
        'stoch_pop':'Stochastic Pop',
        'rsi':'Relative Strength Index',
        'cci':'Commodity Channel Index',
        'momentum':'Momentum',
        'volatility':'Volatility'
        },

    'df_exit_signal_labels':{
        'sar':'Parabolic SAR',
        'sup_res':'Support / Resistance',
        'rsi_trail':'Trailing Relative Strength Index',
        'key_reversal':'Key Reversal Day',
        'trailing_stop':'Trailing Stop',
        'volatility':'Volatility',
        'stoch_cross':'Stochastic Crossover',
        'profit_target':'Profit Target',
        'nday_range':'High/Low Range',
        'random':'Random'
        },

    'df_stop_signal_labels':{
        'initial_dollar':'Initial Dollar',
        'sup_res':'Support / Resistance',
        'immediate_profit':'Immediate Profit',
        'breakeven':'Breakeven',
        'trail_close':'Trailing Close Stop',
        'trail_high_low':'Trailing High / Low Stop'
        },

    # Signal parameter values differing from standard defaults
    'df_entry_signal_dict':{
        '2ma':{
            'ma1':9,
            'ma2':18
            },
        '3ma':{
            'ma1':4,
            'ma2':9,
            'ma3':18
            },
        '4ma':{
            'ma1':5,
            'ma2':12,
            'ma3':20,
            'ma4':40
            },
        'sar':{
            'entry_acceleration_factor':0.02
            },
        'channel_breakout':{
            'entry_period':10
            },
        'stoch_cross':{
            'entry_period':14,
            'entry_oversold':25,
            'entry_overbought':75
            },
        'stoch_over_under':{
            'entry_period':14,
            'entry_oversold':25,
            'entry_overbought':75
            },
        'stoch_pop':{
            'entry_period':14,
            'entry_oversold':25,
            'entry_overbought':75
            },
        'rsi':{
            'entry_period':14,
            'entry_oversold':30,
            'entry_overbought':70
            },
        'cci':{
            'entry_period':20,
            'entry_threshold':0,
            'entry_oversold':-100,
            'entry_overbought':100
            },
        'momentum':{
            'entry_period':10,
            'entry_threshold':0
            },
        'volatility':{
            'entry_period':10,
            'entry_threshold':1.5
            }
        },

    'df_exit_signal_dict':{
        'sar':{
            'exit_period':5,
            'exit_acceleration_factor':0.02,
            'sip_price':True
            },
        'sup_res':{
            'exit_period':5
            },
        'rsi_trail':{
            'exit_period':9,
            'exit_oversold':25,
            'entry_overbought':75
            },
        'key_reversal':{
            'exit_period':10
            },
        'trailing_stop':{
            'exit_amount':3000.00
            },
        'volatility':{
            'exit_period':5,
            'exit_threshold':1
            },
        'stoch_cross':{
            'exit_period':14,
            'exit_oversold':25,
            'entry_overbought':75
            },
        'profit_target':{
            'exit_amount':3000.00
            },
        'nday_range':{
            'exit_period':10
            }
        },

    'df_stop_signal_dict':{
        'initial_dollar':{
            'stop_amount':1500.00
            },
        'sup_res':{
            'stop_period':20
            },
        'immediate_profit':{
            'stop_period':5
            },
        'breakeven':{
            'stop_amount':1500.00
            },
        'trail_close':{
            'stop_amount':1500.00
            },
        'trail_high_low':{
            'stop_amount':1500.00
            }
        },

    'df_params':{
        'equity':100000.00,
        'ticker':'$SPX',
        'bench_ticker':'$SPX',
        'ccy_1':'GBP',
        'ccy_2':'USD',
        'asset_type':'fx',
        'start_date':None,
        'end_date':None,
        'lookback':750,
        'ma1':5,
        'ma2':12,
        'ma3':20,
        'ma4':40,
        'simple_ma':True,
        'fixed_pos_size':5,
        'atr_pos_size':14,
        'position_type':'atr',
        'ticker_source':'norgate',
        'bench_source':'norgate',
        'slippage':5.0,
        'commission':0.00,
        'riskfree':0.0025,
        'entry_type':'2ma',
        'exit_type':'trailing_stop',
        'stop_type':'initial_dollar',
        'entry_period':14,
        'exit_period':5,
        'stop_period':5,
        'entry_threshold':0,
        'exit_threshold':0,
        'exit_amount':3000.00,
        'stop_amount':1500.00,
        'entry_oversold':25,
        'entry_overbought':75,
        'exit_oversold':25,
        'exit_overbought':75,
        'entry_acceleration_factor':0.02,
        'exit_acceleration_factor':0.02,
        'sip_price':False,
        'signals':True,
        'api_key':'',
        'position_risk_bps':500,
        'equity_inv_perc':0.75,
        'margin_%':0.2,
        'contract_months':{
            '01':'F',
            '02':'G',
            '03':'H',
            '04':'J',
            '05':'K',
            '06':'M',
            '07':'N',
            '08':'Q',
            '09':'U',
            '10':'V',
            '11':'X',
            '12':'Z'
        }},

    'df_contract_months':{
        '01':'F',
        '02':'G',
        '03':'H',
        '04':'J',
        '05':'K',
        '06':'M',
        '07':'N',
        '08':'Q',
        '09':'U',
        '10':'V',
        '11':'X',
        '12':'Z'
        }
    }



import pathlib
import runpy
from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent

version_meta = runpy.run_path("./version.py")
VERSION = version_meta["__version__"]
PACKAGE_NAME = 'tradingsystems'
AUTHOR = 'GBERESEARCH'
AUTHOR_EMAIL = 'gberesearch@gmail.com'
URL = 'https://github.com/GBERESEARCH/tradingsystems'

LICENSE = 'MIT'
DESCRIPTION = 'End of day backtesting of Technical trading rules'
LONG_DESCRIPTION = (HERE / "README.md").read_text()
LONG_DESC_TYPE = "text/markdown"

def parse_requirements(filename):
    """Load requirements from a pip requirements file."""
    lineiter = (line.strip() for line in open(filename))
    return [line for line in lineiter if line and not line.startswith("#")]

setup(name=PACKAGE_NAME,
      version=VERSION,
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      long_description_content_type=LONG_DESC_TYPE,
      author=AUTHOR,
      license=LICENSE,
      author_email=AUTHOR_EMAIL,
      url=URL,
      install_requires=parse_requirements("requirements.txt"),
      packages=find_packages()
      )



"""
Profit and Loss functions

"""
# pylint: disable=E1101
import numpy as np
import pandas as pd
pd.options.mode.chained_assignment = None

class Profit():
    """
    Functions for calculating trade profits, mark to market equity, trade
    runs etc.

    """

    @classmethod
    def profit_data(cls, prices, params):
        """
        Adds profit fields to the OHLC data

        Parameters
        ----------
        prices : DataFrame
            The OHLC data.
        position_size : Int, optional
            The number of units to trade. The default is 100.
        slippage : Float, optional
            The amount of slippage to apply to traded prices in basis points.
            The default is 5 bps per unit.
        commission : Float, optional
            The amount of commission charge to apply to each trade. The
            default is $0.00.
        equity : Float
            The initial account equity level.

        Returns
        -------
        prices : DataFrame
            Returns the input DataFrame with additional columns.

        """

        # Create pnl data
        prices = cls._pnl_mtm(
            prices=prices, params=params)

        # Create cumulative trade pnl, equity and drawdown data
        prices = cls._cumulative_trade_pnl_and_equity(
            prices=prices, equity=params['equity'])

        # Create perfect profit data
        prices = cls._perfect_profit(prices=prices, params=params)

        # Calculate margin utilisation
        prices = cls._margin_calc(prices=prices, params=params)

        return prices


    @classmethod
    def _pnl_mtm(cls, prices, params):
        """
        Calculate pnl and mark to market columns

        Parameters
        ----------
        prices : DataFrame
            The OHLC data and trades signals.
        slippage : Float, optional
            The amount of slippage to apply to traded prices in basis points.
            The default is 5 bps per unit.
        commission : Float, optional
            The amount of commission charge to apply to each trade. The
            default is $0.00.

        Returns
        -------
        prices : DataFrame
            The input data with additional columns.

        """

        # daily pnl
        prices = cls._daily_pnl(prices=prices, params=params)

        # total pnl
        prices['total_pnl'] = np.array([0]*len(prices['Close']), dtype=float)
        prices['total_pnl'] = prices['daily_pnl'].cumsum()

        # position mtm
        prices['position_mtm'] = (prices['end_of_day_position']
                                  * prices['Close']
                                  * params['contract_point_value'])

        return prices


    @staticmethod
    def _daily_pnl(prices, params):
        """
        Calculate daily PNL

        Parameters
        ----------
        prices : DataFrame
            The OHLC data and trades signals.
        slippage : Float, optional
            The amount of slippage to apply to traded prices in basis points.
            The default is 5 bps per unit.
        commission : Float, optional
            The amount of commission charge to apply to each trade. The
            default is $0.00.

        Returns
        -------
        prices : DataFrame
            The input data with additional columns.

        """

        # Create series of open, close and position
        open_ = prices['Open']
        close = prices['Close']
        pos = prices['end_of_day_position']

        # Create array of zeros
        day_pnl = np.array([0]*len(close), dtype=float)
        last_day_trade_pnl = np.array([0]*len(close), dtype=float)

        # For each row of closing prices
        for row in range(1, len(close)):

            # If the current position is flat
            if pos[row] == 0:

                # If the previous days position was flat
                if pos[row - 1] == 0:

                    # Set the pnl to zero
                    day_pnl[row] = 0

                # Otherwise:
                else:
                    # Set the pnl to the previous day's position multiplied by
                    # the difference between todays open and yesterdays close
                    # less the cost of slippage and commission
                    day_pnl[row] = (
                        ((pos[row - 1] * (open_[row] - close[row - 1])
                         - abs(pos[row - 1]
                               * params['slippage']
                               * 0.0001
                               * open_[row]))
                        - params['commission'])
                        * params['contract_point_value'])

            # If the current position is not flat
            else:
                # If the position is the same as the previous day
                if pos[row] == pos[row - 1]:

                    # Set the pnl to the current position * the difference
                    # between todays close and yesterdays close
                    day_pnl[row] = (pos[row]
                                    * (close[row] - close[row - 1])
                                    * params['contract_point_value'])

                # If the position is reversed from the previous day
                elif pos[row] == (-1) * pos[row - 1]:
                    day_pnl[row] = (
                        ((pos[row] * (close[row] - open_[row])
                         - abs(pos[row]
                               * params['slippage']
                               * 0.0001
                               * open_[row]))
                        - params['commission'])
                        * params['contract_point_value'])

                    last_day_trade_pnl[row] = (
                        ((pos[row - 1] * (open_[row] - close[row - 1])
                         - abs(
                             pos[row - 1]
                             * params['slippage']
                             * 0.0001
                             * open_[row]))
                        - params['commission'])
                        * params['contract_point_value'])

                # If the position was opened from flat
                else:
                    # Set the pnl to the current position * the difference
                    # between todays open and todays close less the cost of
                    # slippage and commission
                    day_pnl[row] = (
                        ((pos[row] * (close[row] - open_[row])
                         - abs(pos[row]
                               * params['slippage']
                               * 0.0001
                               * open_[row]))
                        - params['commission'])
                        * params['contract_point_value'])

        # Create daily pnl column in DataFrame, rounding to 2dp
        prices['current_trade_pnl'] = np.round(day_pnl, 2)
        prices['last_day_trade_pnl'] = last_day_trade_pnl
        prices['daily_pnl'] = (prices['current_trade_pnl']
                               + prices['last_day_trade_pnl'])

        return prices


    @classmethod
    def _cumulative_trade_pnl_and_equity(cls, prices, equity):
        """
        Calculate cumulative per trade pnl and various account equity series

        Parameters
        ----------
        prices : DataFrame
            The OHLC data and trades signals.
        equity : Float
            The initial account equity level.

        Returns
        -------
        prices : DataFrame
            The input data with additional columns.

        """

        prices = cls._pnl_equity(prices=prices, equity=equity)

        prices = cls._min_max_equity(prices=prices, equity=equity)

        prices = cls._max_dd_gain(prices=prices)

        prices = cls._trade_drawback(prices=prices)

        return prices


    @staticmethod
    def _pnl_equity(prices, equity):
        """
        Calculate cumulative per trade pnl and open, closed and mtm equity
        series

        Parameters
        ----------
        prices : DataFrame
            The OHLC data and trades signals.
        equity : Float
            The initial account equity level.

        Returns
        -------
        prices : DataFrame
            The input data with additional columns.

        """
        # Take the trade number and daily pnl series from prices
        trade_number = prices['trade_number']
        current_trade_pnl = prices['current_trade_pnl']
        last_day_trade_pnl = prices['last_day_trade_pnl']
        daily_pnl = prices['daily_pnl']

        # Create arrays of zeros
        cumulative_trade_pnl = np.array([0.0]*len(daily_pnl))
        mtm_equity = np.array([0.0]*len(daily_pnl))
        closed_equity = np.array([0.0]*len(daily_pnl))
        open_equity = np.array([0.0]*len(daily_pnl))
        max_trade_pnl = np.array([0.0]*len(daily_pnl))

        # Set the initial equity values
        mtm_equity[0] = equity
        closed_equity[0] = equity

        # For each row of data
        for row in range(1, len(daily_pnl)):

            # The index location of the trade entry date
            trade_first_row = prices.index.get_loc(
                prices[trade_number==trade_number[row]].index[0])

            # The number of days since trade entry
            trade_row_num = row - trade_first_row

            # The index location of the trade entry date
            trade_last_row = prices.index.get_loc(
                prices[trade_number==trade_number[row]].index[-1])

            # Set the mtm equity to the previous days mtm equity plus
            # the days pnl
            mtm_equity[row] = mtm_equity[row-1] + daily_pnl[row]

            # If there is a current trade
            if trade_number[row] != 0:

                # If it is the trade entry date and there was no prior trade
                if (trade_row_num == 0
                    and trade_number[row-1] == 0):

                    # Set cumulative trade pnl to the days pnl
                    cumulative_trade_pnl[row] = daily_pnl[row]

                    # The maximum of the initial days pnl and zero
                    max_trade_pnl[row] = max(daily_pnl[row], 0)

                    # Set the closed equity to the previous days closed equity
                    closed_equity[row] = closed_equity[row-1]


                # If it is the trade entry date and this reverses the position
                # of a prior trade
                elif (trade_row_num == 0
                    and trade_number[row-1] != 0):

                    # Set cumulative trade pnl to the previous days cumulative
                    # pnl plus the last days pnl
                    cumulative_trade_pnl[row] = (cumulative_trade_pnl[row-1]
                                                 + last_day_trade_pnl[row])

                    #  Set cumulative trade pnl to the previous days cumulative
                    # pnl plus the last days pnl
                    max_trade_pnl[row] = max(
                        cumulative_trade_pnl[row], max_trade_pnl[row-1])

                    # Set the closed equity to the previous days closed equity
                    closed_equity[row] = (mtm_equity[row-1]
                                          + last_day_trade_pnl[row])


                # If it is the trade exit date and not a reversal
                elif (trade_last_row - row == 0
                      and trade_number[row] == trade_number[row-1]):

                    # Set cumulative trade pnl to the previous days cumulative
                    # pnl plus the days pnl
                    cumulative_trade_pnl[row] = (cumulative_trade_pnl[row-1]
                                                 + daily_pnl[row])

                    # The maximum of the current trade equity and the maximum
                    # trade equity of the previous day
                    max_trade_pnl[row] = max(
                        cumulative_trade_pnl[row], max_trade_pnl[row-1])

                    # Set the closed equity to the mtm equity
                    closed_equity[row] = mtm_equity[row]


                # If it is the second day of a reversal trade
                elif (trade_row_num == 1
                      and trade_number[row-1] != trade_number[row-2]):

                    # Set cumulative trade pnl to the previous days current
                    # pnl plus the days pnl
                    cumulative_trade_pnl[row] = (current_trade_pnl[row-1]
                                                 + daily_pnl[row])

                    # The maximum of the first and second days pnl and zero
                    max_trade_pnl[row] = max(cumulative_trade_pnl[row],
                                             current_trade_pnl[row-1])

                    # Set the closed equity to the previous days closed equity
                    closed_equity[row] = closed_equity[row-1]


                # For every other day in the trade
                else:
                    # Set cumulative trade pnl to the previous days cumulative
                    # pnl plus the days pnl
                    cumulative_trade_pnl[row] = (cumulative_trade_pnl[row-1]
                                                 + daily_pnl[row])

                    # The maximum of the current trade equity and the maximum
                    # trade equity of the previous day
                    max_trade_pnl[row] = max(
                        cumulative_trade_pnl[row], max_trade_pnl[row-1])

                    # Set the closed equity to the previous days closed equity
                    closed_equity[row] = closed_equity[row-1]


            # If there is no current trade
            else:
                # Set cumulative trade pnl to zero
                cumulative_trade_pnl[row] = 0

                # Set the closed equity to the previous days closed equity
                closed_equity[row] = closed_equity[row-1]

            # Current open equity
            open_equity[row] = mtm_equity[row] - closed_equity[row]

        prices['cumulative_trade_pnl'] = cumulative_trade_pnl
        prices['max_trade_pnl'] = max_trade_pnl
        prices['mtm_equity'] = mtm_equity
        prices['closed_equity'] = closed_equity
        prices['open_equity'] = open_equity

        return prices


    @staticmethod
    def _min_max_equity(prices, equity):
        """
        Calculate min and max mtm equity, closed equity, max retracement and
        ulcer index input series

        Parameters
        ----------
        prices : DataFrame
            The OHLC data and trades signals.
        equity : Float
            The initial account equity level.

        Returns
        -------
        prices : DataFrame
            The input data with additional columns.

        """
        # Extract various series from prices
        daily_pnl = prices['daily_pnl']
        mtm_equity = prices['mtm_equity']
        closed_equity = prices['closed_equity']

        # Create arrays of zeros
        max_mtm_equity = np.array([0.0]*len(daily_pnl))
        min_mtm_equity = np.array([0.0]*len(daily_pnl))
        max_closed_equity = np.array([0.0]*len(daily_pnl))
        max_retracement = np.array([0.0]*len(daily_pnl))
        ulcer_index_d_sq = np.array([0.0]*len(daily_pnl))

        # Set the initial equity values
        max_mtm_equity[0] = equity
        min_mtm_equity[0] = equity
        max_closed_equity[0] = equity

        # For each row of data
        for row in range(1, len(daily_pnl)):

            # Maximum mtm equity to this point
            max_mtm_equity[row] = np.max(mtm_equity[0:row+1])

            # Minimum mtm equity to this point
            min_mtm_equity[row] = np.min(mtm_equity[0:row+1])

            # Maximum closed equity to this point
            max_closed_equity[row] = np.max(closed_equity[:row+1])

            # Maximum of max closed equity and current mtm equity, used in
            # calculating Average Max Retracement
            max_retracement[row] = max(
                (max_closed_equity[row] - mtm_equity[row]), 0)

            # Squared difference between max mtm equity and current mtm equity,
            # used in calculating Ulcer Index
            ulcer_index_d_sq[row] = (
                (((max_mtm_equity[row] - mtm_equity[row])
                 / max_mtm_equity[row]) * 100) ** 2)

        prices['max_closed_equity'] = max_closed_equity
        prices['max_retracement'] = max_retracement
        prices['max_mtm_equity'] = max_mtm_equity
        prices['min_mtm_equity'] = min_mtm_equity
        prices['ulcer_index_d_sq'] = ulcer_index_d_sq

        return prices


    @staticmethod
    def _max_dd_gain(prices):
        """
        Calculate max drawdown and max equity gain

        Parameters
        ----------
        prices : DataFrame
            The OHLC data and trades signals.

        Returns
        -------
        prices : DataFrame
            The input data with additional columns.

        """
        # Extract various series from prices
        daily_pnl = prices['daily_pnl']
        mtm_equity = prices['mtm_equity']
        max_mtm_equity = prices['max_mtm_equity']
        min_mtm_equity = prices['min_mtm_equity']

        # Create max drawdown and max gain numpy arrays of zeros
        max_drawdown = np.array([0.0]*len(daily_pnl))
        max_drawdown_perc = np.array([0.0]*len(daily_pnl))
        max_gain = np.array([0.0]*len(daily_pnl))
        max_gain_perc = np.array([0.0]*len(daily_pnl))

        # For each row of data
        for row in range(1, len(daily_pnl)):

            # Maximum drawdown is the smallest value of the current cumulative
            # pnl less the max of all previous rows cumulative pnl and zero
            max_drawdown[row] = mtm_equity[row] - max_mtm_equity[row]

            # Percentage Maximum drawdown
            max_drawdown_perc[row] = ((mtm_equity[row] - max_mtm_equity[row])
                                      / max_mtm_equity[row])

            # Maximum gain is the largest value of the current cumulative
            # pnl less the min of all previous rows and zero
            max_gain[row] = mtm_equity[row] - min_mtm_equity[row]

            # Percentage Maximum gain
            max_gain_perc[row] = ((mtm_equity[row] - min_mtm_equity[row])
                                  / min_mtm_equity[row])

        prices['max_dd'] = max_drawdown
        prices['max_dd_perc'] = max_drawdown_perc
        prices['max_gain'] = max_gain
        prices['max_gain_perc'] = max_gain_perc

        return prices


    @staticmethod
    def _trade_drawback(prices):
        """
        Calculate max trade pnl drawback

        Parameters
        ----------
        prices : DataFrame
            The OHLC data and trades signals.

        Returns
        -------
        prices : DataFrame
            The input data with additional columns.

        """
        # Extract various series from prices
        daily_pnl = prices['daily_pnl']
        cumulative_trade_pnl = prices['cumulative_trade_pnl']
        max_trade_pnl = prices['max_trade_pnl']

        # Create arrays of zeros
        trade_pnl_drawback = np.array([0.0]*len(daily_pnl))
        trade_pnl_drawback_perc = np.array([0.0]*len(daily_pnl))

        # For each row of data
        for row in range(1, len(daily_pnl)):

            # The difference between the highest equity peak of the trade and
            # the current trade open equity
            trade_pnl_drawback[row] = (max_trade_pnl[row]
                                       - cumulative_trade_pnl[row])

            # The percentage difference between the highest equity peak of the
            # trade and the current trade open equity
            if max_trade_pnl[row] !=0:
                trade_pnl_drawback_perc[row] = (
                    (max_trade_pnl[row] - cumulative_trade_pnl[row])
                    / max_trade_pnl[row])

        prices['trade_pnl_drawback'] = trade_pnl_drawback
        prices['trade_pnl_drawback_perc'] = trade_pnl_drawback_perc

        return prices


    @staticmethod
    def _perfect_profit(prices, params):
        """
        Theoretical optimal of buying every low and selling every high

        Parameters
        ----------
        prices : DataFrame
            The OHLC data, trades signals and pnl.
        position_size : Int
            Number of units traded.

        Returns
        -------
        prices : DataFrame
            The input data with additional columns.

        """

        dpp = np.array([0.0]*len(prices))

        for row in range(params['first_trade_start'], len(dpp)):

            # Calculate Daily Perfect Profit
            dpp[row] = (
                abs(prices['High'][row] - prices['Low'][row])
                * prices['position_size_pp'][row])

            # If the High and Low are the same
            if dpp[row] == 0:

                # Use the previous close
                dpp[row] = (
                    abs(prices['High'][row] - prices['Close'][row-1])
                    * prices['position_size_pp'][row])

        # Set this to the daily perfect profit
        prices['daily_perfect_profit'] = dpp * params['contract_point_value']

        # Create array of zeros
        prices['total_perfect_profit'] = np.array(
            [0.0]*len(prices['Close']), dtype=float)

        # Cumulative sum of daily perfect profit column
        prices['total_perfect_profit'] = prices[
            'daily_perfect_profit'].cumsum()

        return prices


    @staticmethod
    def _margin_calc(prices, params):

        prices['total_margin'] = np.array([0.0]*len(prices))

        if params['ticker'][0] == '&':
            prices['initial_margin'] = (
                prices['position_size'] * params['per_contract_margin'])

        else:
            prices['initial_margin'] = (
                prices['Close'] * prices['position_size'] * params['margin_%'])

        for row in range(params['first_trade_start'], len(prices)):
            prices['total_margin'][row] = (
                prices['initial_margin'][row]
                + max(0, -prices['cumulative_trade_pnl'][row]))

        return prices


    @staticmethod
    def time_to_recover(prices):
        """
        Calculate the time taken for largest loss from peak to trough

        Parameters
        ----------
        prices : DataFrame
            The core OHLC DataFrame.

        Returns
        -------
        dd_length : Int
            The number of days from peak to trough.

        """
        # Extract the largest drawdown
        max_dd_val = min(prices['max_dd'])

        # Find the start location of the max drawdown
        dd_start = prices.index.get_loc(
            prices[prices['max_dd']==max_dd_val].index[0])

        # Calculate the point where equity has returned to the pre drawdown
        # level
        if max(prices['max_dd'][dd_start:]) == 0:
            dd_length = (prices['max_dd'][dd_start:].values == 0).argmax()

        # Otherwise return N/A
        else:
            dd_length = 'N/A'

        return dd_length


    @staticmethod
    def time_max_gain(prices):
        """
        Calculate the time taken for largest gain from trough to peak

        Parameters
        ----------
        prices : DataFrame
            The core OHLC DataFrame.

        Returns
        -------
        gain_length : Int
            The number of days from trough to peak.

        """
        # Extract the largest pnl gain
        max_gain_val = max(prices['max_gain'])

        # Reverse the max gain column in time
        gain_rev = prices['max_gain'][::-1]

        # Find the location of the max gain
        max_gain_loc = gain_rev.index.get_loc(
            gain_rev[gain_rev==max_gain_val].index[0])

        # Calculate the time taken from zero to this max gain
        gain_length = (gain_rev[max_gain_loc:].values == 0).argmax()

        return gain_length


    @staticmethod
    def create_monthly_data(prices, equity):
        """
        Create monthly summary data

        Parameters
        ----------
        prices : DataFrame
            The OHLC data.
        equity : Int
            The initial equity level.

        Returns
        -------
        monthly_data : DataFrame
            The monthly summary data.

        """
        # Create empty DataFrame
        monthly_data = pd.DataFrame()

        # Summarize daily pnl data by resampling to monthly
        monthly_data['total_net_profit'] = prices[
            'daily_pnl'].resample('1M').sum()
        monthly_data['average_net_profit'] = prices[
            'daily_pnl'].resample('1M').mean()
        monthly_data['max_net_profit'] = prices[
            'daily_pnl'].resample('1M').max()
        monthly_data['min_net_profit'] = prices[
            'daily_pnl'].resample('1M').min()

        # Create arrays of zeros to hold data
        monthly_data['beginning_equity'] = np.array([0.0]*len(monthly_data))
        monthly_data['additions'] = np.array([0.0]*len(monthly_data))
        monthly_data['withdrawals'] = np.array([0.0]*len(monthly_data))
        monthly_data['end_equity'] = np.array([0.0]*len(monthly_data))
        monthly_data['return'] = np.array([0.0]*len(monthly_data))
        monthly_data['beginning_equity_raw'] = np.array(
            [0.0]*len(monthly_data))
        monthly_data['end_equity_raw'] = np.array([0.0]*len(monthly_data))
        monthly_data['return_raw'] = np.array([0.0]*len(monthly_data))
        monthly_data['abs_loss'] = np.array([0.0]*len(monthly_data))
        monthly_data['abs_loss_raw'] = np.array([0.0]*len(monthly_data))

        # Set initial values
        monthly_data['additions'].iloc[0] = equity
        monthly_data['end_equity'].iloc[0] = (
            monthly_data['beginning_equity'].iloc[0]
            + monthly_data['additions'].iloc[0]
            + monthly_data['withdrawals'].iloc[0]
            + monthly_data['total_net_profit'].iloc[0])
        monthly_data['return'].iloc[0] = (
            (monthly_data['total_net_profit'].iloc[0])
            / (monthly_data['beginning_equity'].iloc[0]
               + monthly_data['additions'].iloc[0]
               + monthly_data['withdrawals'].iloc[0]))
        monthly_data['beginning_equity_raw'].iloc[0] = equity
        monthly_data['end_equity_raw'].iloc[0] = (
            monthly_data['beginning_equity_raw'].iloc[0]
            + monthly_data['total_net_profit'].iloc[0])
        monthly_data['return_raw'].iloc[0] = (
            (monthly_data['total_net_profit'].iloc[0])
            / (monthly_data['beginning_equity_raw'].iloc[0]))

        # For each month
        for row in range(1, len(monthly_data)):

            # Beginning equity is the closing equity from the prior period
            # Raw data keeps all the profits invested whereas the other resets
            # the equity balance each year
            monthly_data['beginning_equity'][row] = monthly_data[
                'end_equity'][row-1]
            monthly_data['beginning_equity_raw'][row] = monthly_data[
                'end_equity_raw'][row-1]

            # For each change in year
            if monthly_data.index.year[row] != monthly_data.index.year[row-1]:

                # If the end of year equity level is less than the initial
                # level
                if monthly_data['end_equity'][row-1] < equity:

                    # Add back the difference to additions
                    monthly_data['additions'][row] = (
                        equity - monthly_data['end_equity'][row-1])
                else:
                    # Otherwise subtract from withdrawals
                    monthly_data['withdrawals'][row] = (
                        equity - monthly_data['end_equity'][row-1])

            # Ending equity is the beginning equity plus the sum of additions,
            # withdrawals and the net pnl over the period
            monthly_data['end_equity'][row] = (
                monthly_data['beginning_equity'][row]
                + monthly_data['total_net_profit'][row]
                + monthly_data['additions'][row]
                + monthly_data['withdrawals'][row])

            # Ending equity raw is the beginning equity plus the net pnl over
            # the period
            monthly_data['end_equity_raw'][row] = (
                monthly_data['beginning_equity_raw'][row]
                + monthly_data['total_net_profit'][row])

            # Monthly return is the net pnl over the period divided by the
            # beginning equity plus the sum of additions and withdrawals
            monthly_data['return'][row] = (
                (monthly_data['total_net_profit'][row])
                / (monthly_data['beginning_equity'][row]
                   + monthly_data['additions'][row]
                   + monthly_data['withdrawals'][row]))

            # Monthly return raw is the net pnl over the period divided by the
            # beginning equity raw
            monthly_data['return_raw'][row] = (
                (monthly_data['total_net_profit'][row])
                / (monthly_data['beginning_equity_raw'][row]))

            # For use in Gain to Pain Ratio, absolute loss is the positive
            # value of the negative monthly returns
            if monthly_data['return'][row] < 0:
                monthly_data['abs_loss'][row] = -monthly_data['return'][row]

            # For use in Gain to Pain Ratio, absolute loss is the positive
            # value of the negative monthly returns
            if monthly_data['return_raw'][row] < 0:
                monthly_data['abs_loss_raw'][row] = -monthly_data[
                    'return_raw'][row]

        return monthly_data



"""
Reporting tools for the trading system data

"""

import datetime as dt
from decimal import Decimal
import numpy as np
from scipy.stats import skew, kurtosis
from tradingsystems.pnl import Profit
from tradingsystems.winloss import Runs


class PerfReport():
    """
    Create a dictionary of performance data and produce a text report of the
    system performance.

    """

    @classmethod
    def performance_data(cls, tables, params, labels, norgate_name_dict):
        """
        Create dictionary of performance data.

        Parameters
        ----------
        prices : DataFrame
            The OHLC data, trades signals and pnl.
        monthly_data : DataFrame
            The monthly summary data.
        ticker_source : Str
            The data source to use for the ticker data, either 'norgate',
            'alpha' or 'yahoo'. The default is 'norgate'.
        asset_type : Str
            The alphavantage asset class type. The default is 'fx'.
        ccy_1 : Str
            Primary currency of pair to return. The default 'GBP'.
        ccy_2 : Str
            Secondary currency of pair to return. The default 'USD'.
        ticker : Str
            Underlying to test. The default '$SPX'.
        entry_label : Str
            The longname of the entry strategy.
        exit_label : Str
            The longname of the exit strategy.
        stop_label : Str
            The longname of the stop strategy.
        norgate_name_dict : Dict
            Dictionary lookup of Norgate tickers to long names.
        slippage : Float
            The amount of slippage to apply to traded prices in basis points.
            The default is 5 bps per unit.
        commission : Float
            The amount of commission charge to apply to each trade. The
            default is $0.00.
        position_size : Int
            The number of units to trade. The default is based on equity.
        benchmark : Series
            The closing prices of the benchmark.
        benchmark_position_size : Int
            The number of units of the benchmark to trade. The default is based
            on equity.
        riskfree : Float, optional
            The riskfree interest rate. The default is 25bps.

        Returns
        -------
        perf_dict : Dict
            Dictionary of performance data.

        """

        # Create empty dictionary
        perf_dict = {}

        prices = tables['prices']

        perf_dict = cls._perf_data_init(
            perf_dict=perf_dict, prices=prices, params=params, labels=labels,
            norgate_name_dict=norgate_name_dict)

        # Maximum margin required
        perf_dict['max_init_margin'] = max(prices['initial_margin'])
        perf_dict['max_total_margin'] = max(prices['total_margin'])

        # Maximum number of contracts held
        perf_dict['max_contracts'] = abs(max(prices['end_of_day_position']))

        # Net Profit
        perf_dict['net_pnl'] = prices['total_pnl'].iloc[-1]

        # Convert the start and end date strings to dates
        perf_dict['start'] = dt.datetime.strptime(
            perf_dict['start_date'], "%d/%m/%y")
        perf_dict['end'] = dt.datetime.strptime(
            perf_dict['end_date'], "%d/%m/%y")

        # Number of days between start and end dates
        perf_dict['period'] = (perf_dict['end']-perf_dict['start']).days

        # Set the return period in years
        perf_dict['return_period'] = perf_dict['period'] / 365

        # Annualized profit
        perf_dict['annualized_profit '] = (
            perf_dict['net_pnl'] / perf_dict['return_period'])

        # Total return
        perf_dict['total_return_rate'] = (
            perf_dict['net_pnl'] / perf_dict['initial_equity'])

        # Annualized total return
        perf_dict['annualized_return'] = (
            (1 + perf_dict['total_return_rate'])
            ** (1 / perf_dict['return_period']) - 1)

        perf_dict = cls._perf_data_trades(perf_dict=perf_dict, tables=tables)

        # Pessimistic Return on Margin
        perf_dict['adjusted_gross_profit'] = perf_dict[
            'av_win'] * (perf_dict['num_wins']
                         - np.round(np.sqrt(perf_dict['num_wins'])))

        perf_dict['adjusted_gross_loss'] = perf_dict[
            'av_loss'] * (perf_dict['num_losses']
                          - np.round(np.sqrt(perf_dict['num_losses'])))

        perf_dict['prom'] = np.round(
            (perf_dict['adjusted_gross_profit']
             - perf_dict['adjusted_gross_loss']) /
            perf_dict['max_init_margin'], 2)

        # PROM minus biggest win
        perf_dict['prom_minus_max_win'] = np.round(
            (perf_dict['adjusted_gross_profit']
             - perf_dict['max_win']
             - perf_dict['adjusted_gross_loss'])
            / perf_dict['max_init_margin'], 2)

        # Perfect profit - Buy every dip & sell every peak
        perf_dict['perfect_profit'] = prices['total_perfect_profit'].iloc[-1]

        # Model Efficiency - ratio of net pnl to perfect profit
        perf_dict['model_efficiency'] = np.round(
            (perf_dict['net_pnl'] / perf_dict['perfect_profit']) * 100, 2)

        # Winning run data
        perf_dict['win_run_dict'] = Runs.trade_runs(
            perf_dict['trade_data_dict']['trades_win_list'], run_type='win')

        # Losing run data
        perf_dict['loss_run_dict'] = Runs.trade_runs(
            perf_dict['trade_data_dict']['trades_loss_list'], run_type='loss')

        # Maximum Equity drawdown
        perf_dict['max_balance_drawback'] = np.round(min(prices['max_dd']), 2)

        # Maximum Equity drawdown in percentage terms
        perf_dict['max_balance_drawback_perc'] = min(
            prices['max_dd_perc']) * 100

        # Time to recover from Max Drawdown
        perf_dict['time_to_recover'] = Profit.time_to_recover(prices)

        # Maximum Equity gain
        perf_dict['max_gain'] = np.round(max(prices['max_gain']), 2)

        # Maximum Equity gain in percentage terms
        perf_dict['max_gain_perc'] = max(prices['max_gain_perc']) * 100

        # Time taken for maximum gain
        perf_dict['max_gain_time'] = Profit.time_max_gain(prices)

        # Reward / Risk ratio
        perf_dict['reward_risk'] = (perf_dict['net_pnl'] /
                                    abs(perf_dict['max_balance_drawback']))

        # Annual Rate of Return
        perf_dict['annual_ror'] = (perf_dict['annualized_profit '] /
                                   perf_dict['initial_equity']) * 100

        # Profit Index
        perf_dict['profit_factor'] = (perf_dict['total_profit'] /
                                     abs(perf_dict['total_loss']))

        # Mathematical Advantage
        perf_dict['mathematical_advantage'] = np.round(
            (perf_dict['num_wins']
             / len(perf_dict['trade_data_dict']['trades']) * 100)
            * (perf_dict['profit_factor'] + 1) - 100, 2)

        # Get the index of the first trade entry
        # Select just the rows of the first trade from the DataFrame
        #first_trade = prices[prices['trade_number']==1]

        # Find the date of the trade entry
        #trade_start_date = first_trade.index[0]

        # Return of a buy and hold strategy since the first trade entry
        perf_dict['long_only_pnl'] = (
            (prices['Close'].iloc[-1]
             - prices['Open'].loc[params['first_trade_date']]) *
            params['init_position_size']) * params['contract_point_value']

        # Annual Rate of return of buy and hold
        perf_dict['annual_long_only_ror'] = (
            (perf_dict['long_only_pnl'] / perf_dict['return_period']) /
            perf_dict['initial_equity']) * 100

        # Return of a buy and hold strategy of SPX since the first trade entry
        perf_dict['long_only_pnl_spx'] = (
            (tables['benchmark']['Close'].iloc[-1]
             - tables['benchmark']['Open'].loc[params['first_trade_date']]) *
            params['init_benchmark_position_size'])

        # Annual Rate of return of buy and hold of SPX
        perf_dict['annual_long_only_spx_ror'] = (
            (perf_dict['long_only_pnl_spx'] / perf_dict['return_period']) /
            perf_dict['initial_equity']) * 100

        perf_dict = cls._perf_data_stat_measures(
            perf_dict=perf_dict, tables=tables, params=params)

        # Maximum MTM Equity Profit
        perf_dict['max_equity_profit'] = (
            prices['max_mtm_equity'].iloc[-1] - perf_dict['initial_equity'])

        # Largest open equity drawdown
        perf_dict['open_equity_dd'] = (prices['trade_pnl_drawback'].max())*(-1)

        # Open equity
        perf_dict['open_equity'] = prices['open_equity'].iloc[-1]

        # Closed equity
        perf_dict['closed_equity'] = prices['closed_equity'].iloc[-1]

        # Largest monthly gain
        perf_dict['month_net_pnl_large'] = (
            tables['monthly_data']['total_net_profit'].max())

        # Largest monthly loss
        perf_dict['month_net_pnl_small'] = (
            tables['monthly_data']['total_net_profit'].min())

        # Average monthly gain/loss
        perf_dict['month_net_pnl_av'] = (
            tables['monthly_data']['total_net_profit'].mean())

        # Values still to be worked out
        placeholder_dict = {
            #'pessimistic_margin':0.00,
            #'adj_pess_margin':0.00,
            'pess_month_avg':0.00,
            'pess_month_variance':0.00,
            'mod_pess_margin':0.00,
            }

        perf_dict.update(placeholder_dict)

        return perf_dict


    @staticmethod
    def _perf_data_init(perf_dict, prices, params, labels, norgate_name_dict):

        # Contract and strategy details
        if (params['ticker_source'] == 'alpha'
            and params['asset_type'] in ['fx', 'crypto']):
            perf_dict['contract'] = params['ccy_1'] + params['ccy_2']
        else:
            perf_dict['contract'] = params['ticker']

        # Entry, exit and stop labels
        perf_dict['entry_label'] = labels['entry_label']
        perf_dict['exit_label'] = labels['exit_label']
        perf_dict['stop_label'] = labels['stop_label']
        perf_dict['position_size_label'] = params['position_size_label']

        # Initial Equity
        perf_dict['initial_equity'] = prices['mtm_equity'].iloc[0]

        # Set Ticker Longname
        if params['ticker_source'] == 'norgate':
            perf_dict['longname'] = norgate_name_dict[params['ticker']]
        else:
            perf_dict['longname'] = perf_dict['contract']

        # Slippage and commission in dollars
        perf_dict['slippage'] = params['slippage']
        perf_dict['commission'] = params['commission']

        # Start and end dates
        perf_dict['start_date'] = prices.index[0].date().strftime("%d/%m/%y")
        perf_dict['end_date'] = prices.index[-1].date().strftime("%d/%m/%y")

        return perf_dict


    @staticmethod
    def _perf_data_trades(perf_dict, tables):

        prices = tables['prices']

        # Calculate trade data
        perf_dict['trade_data_dict'] = Runs.trade_data(prices)

        perf_dict['total_trades'] = perf_dict['trade_data_dict']['num_trades']

        # number of wins and losses
        perf_dict['num_wins'] = len(
            perf_dict['trade_data_dict']['trades_win_dict'])
        perf_dict['num_losses'] = len(
            perf_dict['trade_data_dict']['trades_loss_dict'])

        # Sum of all profits
        perf_dict['total_profit'] = np.round(
            sum(perf_dict['trade_data_dict']['trades_win_dict'].values()), 2)

        # Sum of all losses
        perf_dict['total_loss'] = np.round(
            sum(perf_dict['trade_data_dict']['trades_loss_dict'].values()), 2)

        # Percentage of winning trades
        perf_dict['win_percent'] = int(
            np.round(perf_dict['num_wins'] / len(
                perf_dict['trade_data_dict']['trades']) * 100, 0))

        # Percentage of losing trades
        perf_dict['loss_percent'] = int(
            np.round(perf_dict['num_losses']
                     / len(perf_dict['trade_data_dict']['trades']) * 100, 0))


        # If there are winning trades
        if perf_dict['trade_data_dict']['trades_win_dict'].values():

            # Max winning trade
            perf_dict['max_win'] = np.round(
                max(perf_dict['trade_data_dict']['trades_win_dict'].values()),
                2)

            # Min winning trade
            perf_dict['min_win'] = np.round(
                min(perf_dict['trade_data_dict']['trades_win_dict'].values()),
                2)

            # Average winning trade
            perf_dict['av_win'] = np.round(
                sum(perf_dict['trade_data_dict']['trades_win_dict'].values())
                / len(perf_dict['trade_data_dict']['trades_win_list']), 2)

            # standard deviation of winning trades
            perf_dict['std_wins'] = np.round(
                np.std(perf_dict['trade_data_dict']['trades_win_list']), 2)

        # Otherwise set these to zero
        else:
            perf_dict['max_win'] = 0
            perf_dict['min_win'] = 0
            perf_dict['av_win'] = 0
            perf_dict['std_wins'] = 0


        # If there are losing trades
        if perf_dict['trade_data_dict']['trades_loss_dict'].values():

            # Max losing trade
            perf_dict['max_loss'] = min(
                perf_dict['trade_data_dict']['trades_loss_dict'].values())

            # Min losing trade
            perf_dict['min_loss'] = max(
                perf_dict['trade_data_dict']['trades_loss_dict'].values())

            # average losing trade
            perf_dict['av_loss'] = np.round(
                sum(perf_dict['trade_data_dict']['trades_loss_dict'].values())
                / len(perf_dict['trade_data_dict']['trades_loss_list']), 2)

            # standard deviation of losing trades
            perf_dict['std_losses'] = np.round(
                np.std(perf_dict['trade_data_dict']['trades_loss_list']), 2)

        # Otherwise set these to zero
        else:
            perf_dict['max_loss'] = 0
            perf_dict['min_loss'] = 0
            perf_dict['av_loss'] = 0
            perf_dict['std_losses'] = 0


        # Maximum win/loss ratio
        perf_dict['max_win_loss_ratio'] = np.round(
            abs(perf_dict['max_win'] / perf_dict['max_loss']), 2)


        # Minimum win/loss ratio
        perf_dict['min_win_loss_ratio'] = np.round(
            abs(perf_dict['min_win'] / perf_dict['min_loss']), 2)


        # avwin / avloss
        perf_dict['av_win_loss_ratio'] = np.round(
            abs(perf_dict['av_win'] / perf_dict['av_loss']), 2)

        # average trade
        perf_dict['av_trade'] = np.round(
            perf_dict['net_pnl']
            / len(perf_dict['trade_data_dict']['trades']), 2)

        return perf_dict


    @staticmethod
    def _perf_data_stat_measures(perf_dict, tables, params):

        prices = tables['prices']

        # Riskfree Rate
        perf_dict['riskfree_rate'] = params['riskfree']

        # Mean Price
        perf_dict['close_price_mean'] = np.round(np.mean(prices['Close']), 2)

        # Variance Price
        perf_dict['close_price_variance'] = np.round(
            np.var(prices['Close']), 2)

        # Standard Deviation Price
        perf_dict['close_price_std_dev'] = np.round(np.std(prices['Close']), 2)

        # Skewness Price
        perf_dict['close_price_skewness'] = np.round(skew(prices['Close']), 2)

        # Kurtosis Price
        perf_dict['close_price_kurtosis'] = np.round(
            kurtosis(prices['Close']), 2)

        # Mean Return
        perf_dict['close_return_mean'] = np.round(
            np.mean(prices['Close'].pct_change()*100), 2)

        # Variance Return
        perf_dict['close_return_variance'] = np.round(
            np.var(prices['Close'].pct_change()*100), 2)

        # Standard Deviation Return
        perf_dict['close_return_std_dev'] = np.round(
            np.std(prices['Close'].pct_change()*100), 2)

        # Annualized Volatility
        perf_dict['close_return_ann_vol'] = np.round(
            np.std(prices['Close'].pct_change()*100)*np.sqrt(252), 2)

        # Skewness Return
        perf_dict['close_return_skewness'] = np.round(
            skew(prices['Close'].pct_change()*100, nan_policy='omit'), 2)

        # Kurtosis Return
        perf_dict['close_return_kurtosis'] = np.round(kurtosis(
            prices['Close'].pct_change()*100, nan_policy='omit'), 2)

        # Efficiency Ratio
        perf_dict['efficiency_ratio'] = np.round(
            (abs(prices['Close'][-1] - prices['Close'][0])
            / np.nansum(abs(prices['Close'] - prices['Close'].shift())))
            * 100, 2)

        # MTM Equity Annualized Standard Deviation
        perf_dict['equity_std_dev'] = np.round(
            np.std(prices['mtm_equity'].pct_change()*100)*np.sqrt(252), 2)

        # Sharpe Ratio
        perf_dict['sharpe_ratio'] = ((
            perf_dict['annual_ror'] - perf_dict['riskfree_rate'])
            / perf_dict['equity_std_dev'])

        # Information Ratio
        perf_dict['information_ratio'] = (
            perf_dict['annual_ror'] / perf_dict['equity_std_dev'])

        # Treynor Ratio
        perf_dict['treynor_inv_correl'] = tables[
            'benchmark'].Close.pct_change().corr(
                prices.mtm_equity.pct_change())
        #stock_correl = benchmark.Close.pct_change().corr(
        #    prices.Close.pct_change())
        perf_dict['treynor_sd_inv'] = np.std(
            prices.mtm_equity.pct_change())*np.sqrt(252)
        #sd_stock = np.std(prices.Close.pct_change())
        perf_dict['treynor_sd_index'] = np.std(
            tables['benchmark'].Close.pct_change())*np.sqrt(252)
        #beta = stock_correl * (sd_stock / sd_index)
        perf_dict['treynor_beta'] = (perf_dict['treynor_inv_correl']
                                     * (perf_dict['treynor_sd_inv']
                                        / perf_dict['treynor_sd_index']))
        treynor = np.round(
            (perf_dict['annual_ror'] - perf_dict['riskfree_rate'])
            / perf_dict['treynor_beta'], 2)
        if treynor > 0:
            perf_dict['treynor_ratio'] = treynor
        else:
            perf_dict['treynor_ratio'] = 'N/A'

        # Sortino Ratio
        equity_return = prices['mtm_equity'].pct_change()*100
        downside_deviation = np.std(
            equity_return[equity_return < 0])*np.sqrt(252)
        perf_dict['sortino_ratio'] = np.round((
            perf_dict['annual_ror'] - perf_dict['riskfree_rate'])
            / downside_deviation, 2)

        # Calmar Ratio
        perf_dict['calmar_ratio'] = (
            perf_dict['annual_ror']
            / abs(perf_dict['max_balance_drawback_perc']))

        # Average Maximum Retracement
        perf_dict['average_max_retracement'] = (
            prices['max_retracement'].mean())

        # Ulcer Index
        perf_dict['ulcer_index'] = np.sqrt(
            prices['ulcer_index_d_sq'][prices['ulcer_index_d_sq']!=0].mean())

        # Gain to Pain
        perf_dict['gain_to_pain'] = np.round(
            tables['monthly_data']['return'].sum()
            / tables['monthly_data']['abs_loss'].sum(), 2)

        return perf_dict


    @classmethod
    def report_table(cls, input_dict):
        """
        Print out backtesting performance results.

        Modelled on performance results presented in Robert Pardo, 1992,
        Design, Testing and Optimization of Trading Systems

        Parameters
        ----------
        input_dict : Dict
            Dictionary of performance data.

        Returns
        -------
        Prints results to the console.

        """

        win_run_dict = input_dict['win_run_dict']
        loss_run_dict = input_dict['loss_run_dict']

        # Format the performance dictionary so that the financial data is
        # rounded to 2 decimal places and set values as strings
        input_dict = cls._dict_format(input_dict)
        win_run_dict = cls._dict_format(win_run_dict)
        loss_run_dict = cls._dict_format(loss_run_dict)

        # Core strategy details
        cls._report_header(input_dict)

        # Trade Win / Loss statistics
        cls._report_trades(input_dict, win_run_dict, loss_run_dict)

        # Equity return / drawdown statistics
        cls._report_equity(input_dict)

        # Key performance measures
        cls._report_key_perf(input_dict)

        # Data distribution statistics
        cls._report_data_dist(input_dict)


    @staticmethod
    def _report_header(input_dict):

        # Format header - centred and with lines above and below
        print('='*78)
        print('{:^78}'.format('Performance Analysis Report'))
        print('-'*78)

        # Contract traded on left and period covered on right
        print('Contract Ticker : {:<41}{} - {}'.format(
            input_dict['contract'],
            input_dict['start_date'],
            input_dict['end_date']))

        # Contract Longname
        print('Contract Name   : {:>10}'.format(input_dict['longname']))

        # Entry Method
        print('Entry           : {:>10}'.format(input_dict['entry_label']))

        # Exit Method
        print('Exit            : {:>10}'.format(input_dict['exit_label']))

        # Stop Method
        print('Stop            : {:>10}'.format(input_dict['stop_label']))

        # Position size Method
        print('Position Size   : {:>10}'.format(
            input_dict['position_size_label']))

        # Beginning equity balance and max contracts traded
        print('Initial Equity  :     ${:>10}{:>9}{}{:>11}'.format(
            input_dict['initial_equity'],
            '',
            'Max Contracts    :  ',
            input_dict['max_contracts']))

        # Maximum Margin
        print('Max Init Margin :     ${:>10}{:>9}{}{:>11}'.format(
            input_dict['max_init_margin'],
            '',
            'Max Total Margin : $',
            input_dict['max_total_margin']))

        # Commission and slippage
        print('Commission      :     ${:>10}{:>9}{}{:>11}'.format(
            input_dict['commission'],
            '',
            'Slippage         :  ',
            input_dict['slippage']+'bps'))

        # Empty line
        print()

        # Profit stats
        print('Net P/L         :     ${:>10}{:>9}{}{:>11}'.format(
            input_dict['net_pnl'],
            '',
            'Perfect Profit   : $',
            input_dict['perfect_profit']))
        print('Total Profit    :     ${:>10}{:>9}{}{:>11}'.format(
            input_dict['total_profit'],
            '',
            'Total Loss       : $',
            input_dict['total_loss']))
        print('Av. Trade P/L   :     ${:>10}{:>9}{}{:>11}'.format(
            input_dict['av_trade'],
            '',
            'Model Efficiency : %',
            input_dict['model_efficiency']))


    @staticmethod
    def _report_trades(input_dict, win_run_dict, loss_run_dict):

        # Headers for # trades, Max, Min and Average with lines above and below
        print('-'*78)
        print('{:>16}:{:^10}:{:^17}:{:^17}:{:^14}'.format(
            '',
            '# Trades',
            'Maximum',
            'Minimum',
            'Average'))
        print('-'*78)

        # Total trades summary
        print('Total...........:{:^10}:{:^17}:{:^17}:'.format(
            input_dict['total_trades'],
            'Win {} / {}%'.format(
                input_dict['num_wins'], input_dict['win_percent']),
            'Loss {} / {}%'.format(
                input_dict['num_losses'], input_dict['loss_percent'])))

        # Winning trades and winning runs
        print('Win.............:{:^10}:{:>11}{:>6}:{:>10}{:>7}:{:>9}'.format(
            input_dict['num_wins'],
            input_dict['max_win'],
            '',
            input_dict['min_win'],
            '',
            input_dict['av_win']))

        print('Win runs........:{:^10}:{:>11}{:<6}:{:>10}{:<7}:{:>9}{}'.format(
            win_run_dict['num_win_runs'],
            win_run_dict['max_win_run_pnl'],
            ' / {}'.format(win_run_dict['max_win_run_count']),
            win_run_dict['min_win_run_pnl'],
            ' / {}'.format(win_run_dict['min_win_run_count']),
            win_run_dict['av_win_run_pnl'],
            ' / {}'.format(win_run_dict['av_win_run_count'])))

        # Losing trades and losing runs
        print('Loss............:{:^10}:{:>11}{:>6}:{:>10}{:>7}:{:>9}'.format(
            input_dict['num_losses'],
            input_dict['max_loss'],
            '',
            input_dict['min_loss'],
            '',
            input_dict['av_loss']))

        print('Loss runs.......:{:^10}:{:>11}{:<6}:{:>10}{:<7}:{:>9}{}'.format(
            loss_run_dict['num_loss_runs'],
            loss_run_dict['max_loss_run_pnl'],
            ' / {}'.format(loss_run_dict['max_loss_run_count']),
            loss_run_dict['min_loss_run_pnl'],
            ' / {}'.format(loss_run_dict['min_loss_run_count']),
            loss_run_dict['av_loss_run_pnl'],
            ' / {}'.format(loss_run_dict['av_loss_run_count'])))

        # Win loss ratio
        print('Win/Loss Ratio..:{:>10}:{:>11}{:>6}:{:>10}{:>7}:{:>9}'.format(
            '',
            input_dict['max_win_loss_ratio'],
            '',
            input_dict['min_win_loss_ratio'],
            '',
            input_dict['av_win_loss_ratio']))

        # Separating line
        print('-'*78)


    @staticmethod
    def _report_equity(input_dict):

        # Open and Closed equity
        print('Open Equity.......... ${:>10}{:<6}{}{:>10}'.format(
            input_dict['open_equity'],
            '',
            'Closed Equity......... $',
            input_dict['closed_equity']))

        # Max Open Equity DD / equity profit
        print('Max Open Eq. DD...... ${:>10}{:<6}{}{:>10}'.format(
            input_dict['open_equity_dd'],
            '',
            'Max Equity Profit..... $',
            input_dict['max_equity_profit']))

        # Max Closed Equity DD drawback / Max Equity Gain
        print('Peak to Trough DD.... ${:>10}{:<6}{}{:>10}'.format(
            input_dict['max_balance_drawback'],
            '',
            'Trough to Peak Gain... $',
            input_dict['max_gain']))

        # Max Percent drawdown / gain
        print('Max Drawdown......... %{:>10}{:<6}{}{:>10}'.format(
            input_dict['max_balance_drawback_perc'],
            '',
            'Max Gain.............. %',
            input_dict['max_gain_perc']))

        # Time to recover and time for max gain
        print('Time to Recover (Days) {:>10}{:<6}{}{:>10}'.format(
            input_dict['time_to_recover'],
            '',
            'Time for Max Gain (Days)',
            input_dict['max_gain_time']))

        # Reward / Risk
        print('Reward / Risk........  {:^16}{}{:>10}'.format(
            '{} TO 1'.format(input_dict['reward_risk']),
            'Annual Rate of Return. %',
            input_dict['annual_ror']))

        # Profit Index and Mathematical Advantage
        print('Profit Factor........  {:>10}{:<6}{}{:>10}'.format(
            input_dict['profit_factor'],
            '',
            'Mathematical Advantage %',
            input_dict['mathematical_advantage']))

        # Pessimistic Margin
        print('Pessimistic Margin...  {:>10}{:<6}{}{:>10}'.format(
            input_dict['prom'],
            '',
            'Adjusted Pess. Margin.  ',
            input_dict['prom_minus_max_win']))

        print('Pess. Month Avg......  {:>10}{:<6}{}{:>10}'.format(
            input_dict['pess_month_avg'],
            '',
            'Pess. Month Variance..  ',
            input_dict['pess_month_variance']))

        # Monthly P/L
        print('Monthly Net PL Large. ${:>10}{:<6}{}{:>10}'.format(
            input_dict['month_net_pnl_large'],
            '',
            'Monthly Net PL Small.. $',
            input_dict['month_net_pnl_small']))

        print('Monthly Net PL Ave... ${:>10}{:<6}{}{:>10}'.format(
            input_dict['month_net_pnl_av'],
            '',
            'Modified Pess. Margin.  ',
            input_dict['mod_pess_margin']))

        print('Long Only Net PL..... ${:>10}{:<6}{}{:>10}'.format(
            input_dict['long_only_pnl'],
            '',
            'Long Only Annual RoR.. %',
            input_dict['annual_long_only_ror']))

        print('Long Only SPX Net PL. ${:>10}{:<6}{}{:>10}'.format(
            input_dict['long_only_pnl_spx'],
            '',
            'Long Only SPX Ann RoR. %',
            input_dict['annual_long_only_spx_ror']))


    @staticmethod
    def _report_key_perf(input_dict):

        # Key Performance Measures
        print('-'*78)
        print('{:^78}'.format('Key Performance Measures'))
        print('-'*78)

        # Sharpe Ratio & Information Ratio
        print('Sharpe Ratio.........  {:>10}{:<6}{}{:>10}'.format(
            input_dict['sharpe_ratio'],
            '',
            'Information Ratio.....  ',
            input_dict['information_ratio']))

        # Sortino Ratio & Treynor Ratio
        print('Sortino Ratio........  {:>10}{:<6}{}{:>10}'.format(
            input_dict['sortino_ratio'],
            '',
            'Treynor Ratio.........  ',
            input_dict['treynor_ratio']))

        # Calmar Ratio & Average Max Retracement
        print('Calmar Ratio.........  {:>10}{:<6}{}{:>10}'.format(
            input_dict['calmar_ratio'],
            '',
            'Av. Max Retracement... $',
            input_dict['average_max_retracement']))

        # Ulcer Index & Gain to Pain
        print('Ulcer Index..........  {:>10}{:<6}{}{:>10}'.format(
            input_dict['ulcer_index'],
            '',
            'Gain to Pain..........  ',
            input_dict['gain_to_pain']))

        # Price Distribution statistics
        print('-'*78)
        print('{:^78}'.format('Data Distribution Statistics'))
        print('-'*78)

        # Mean & Standard Deviation of Prices
        print('Mean Price........... ${:>10}{:<6}{}{:>10}'.format(
            input_dict['close_price_mean'],
            '',
            'St. Dev. Price........ $',
            input_dict['close_price_std_dev']))

        # Skewness & Kurtosis of Prices
        print('Skewness Price.......  {:>10}{:<6}{}{:>10}'.format(
            input_dict['close_price_skewness'],
            '',
            'Kurtosis Price........  ',
            input_dict['close_price_kurtosis']))


    @staticmethod
    def _report_data_dist(input_dict):

        # Return Distribution statistics

        # Mean & Standard Deviation of Returns
        print('Mean Return.......... %{:>10}{:<6}{}{:>10}'.format(
            input_dict['close_return_mean'],
            '',
            'St. Dev. Return....... %',
            input_dict['close_return_std_dev']))

        # Skewness & Kurtosis of Returns
        print('Skewness Return......  {:>10}{:<6}{}{:>10}'.format(
            input_dict['close_return_skewness'],
            '',
            'Kurtosis Return.......  ',
            input_dict['close_return_kurtosis']))

        # Efficiency Ratio & Annualized Volatility
        print('Efficiency Ratio..... %{:>10}{:<6}{}{:>10}'.format(
            input_dict['efficiency_ratio'],
            '',
            'Annualized Vol........ %',
            input_dict['close_return_ann_vol']))

        # Closing line
        print('='*78)


    @staticmethod
    def _dict_format(input_dict):
        """
        Format the performance dictionary so that the financial data is
        rounded to 2 decimal places and set values as strings

        Parameters
        ----------
        input_dict : Dict
            The performance data dictionary.

        Returns
        -------
        str_input_dict : Dict
            The formatted performance data dictionary.

        """

        # Create empty dictionary
        str_input_dict = {}

        # Set decimal format
        dp2 = Decimal(10) ** -2  # (equivalent to Decimal '0.01')

        # For each entry in the dictionary
        for key, value in input_dict.items():

            # If the value is a floating point number
            if type(value) in (float, np.float64):

                # Apply the decimal formatting and convert to string
                str_input_dict[key] = str(Decimal(value).quantize(dp2))
            else:

                # Otherwise just convert to string
                str_input_dict[key] = str(value)

        return str_input_dict
